import sys
import pytest
from sdl2 import mouse

# NOTE: Will need to use a window fixture to test most of these. Will also want
# to mark most as "SKIP_ANNOYING".


@pytest.mark.skip("not implemented")
def test_SDL_GetMouseFocus():
    pass

@pytest.mark.skip("not implemented")
def test_SDL_GetMouseState():
    pass

@pytest.mark.skip("not implemented")
def test_SDL_GetRelativeMouseState():
    pass

@pytest.mark.skip("not implemented")
def test_SDL_WarpMouseInWindow():
    pass

@pytest.mark.skip("not implemented")
def test_SDL_SetRelativeMouseMode():
    pass

@pytest.mark.skip("not implemented")
def test_SDL_GetRelativeMouseMode():
    pass

@pytest.mark.skip("not implemented")
def test_SDL_CreateCursor():
    pass

@pytest.mark.skip("not implemented")
def test_SDL_CreateColorCursor():
    pass

@pytest.mark.skip("not implemented")
def test_SDL_CreateSystemCursor():
    pass

@pytest.mark.skip("not implemented")
def test_SDL_SetCursor():
    pass

@pytest.mark.skip("not implemented")
def test_SDL_GetCursor():
    pass

@pytest.mark.skip("not implemented")
def test_SDL_GetDefaultCursor():
    pass

@pytest.mark.skip("not implemented")
def test_SDL_FreeCursor():
    pass

@pytest.mark.skip("not implemented")
def test_SDL_ShowCursor():
    pass

@pytest.mark.skip("not implemented")
def test_SDL_WarpMouseGlobal():
    pass

@pytest.mark.skip("not implemented")
def test_SDL_CaptureMouse():
    pass

@pytest.mark.skip("not implemented")
def test_SDL_GetGlobalMouseState():
    pass



from ctypes import Structure, c_int, c_char_p, POINTER
from .dll import _bind
from .stdinc import Uint8, Uint32
from .video import SDL_Window

__all__ = [
    # Structs
    "SDL_MessageBoxButtonData", "SDL_MessageBoxColor",
    "SDL_MessageBoxColorScheme", "SDL_MessageBoxData",

    # Enums
    "SDL_MessageBoxFlags",
    "SDL_MESSAGEBOX_ERROR", "SDL_MESSAGEBOX_WARNING",
    "SDL_MESSAGEBOX_INFORMATION", "SDL_MESSAGEBOX_BUTTONS_LEFT_TO_RIGHT",
    "SDL_MESSAGEBOX_BUTTONS_RIGHT_TO_LEFT",

    "SDL_MessageBoxButtonFlags",
    "SDL_MESSAGEBOX_BUTTON_RETURNKEY_DEFAULT",
    "SDL_MESSAGEBOX_BUTTON_ESCAPEKEY_DEFAULT",

    "SDL_MessageBoxColorType",
    "SDL_MESSAGEBOX_COLOR_BACKGROUND", "SDL_MESSAGEBOX_COLOR_TEXT",
    "SDL_MESSAGEBOX_COLOR_BUTTON_BORDER",
    "SDL_MESSAGEBOX_COLOR_BUTTON_BACKGROUND",
    "SDL_MESSAGEBOX_COLOR_BUTTON_SELECTED", "SDL_MESSAGEBOX_COLOR_MAX",

    # Functions
    "SDL_ShowMessageBox", "SDL_ShowSimpleMessageBox"
]


SDL_MessageBoxFlags = c_int

SDL_MESSAGEBOX_ERROR = 0x00000010
SDL_MESSAGEBOX_WARNING = 0x00000020
SDL_MESSAGEBOX_INFORMATION = 0x00000040
SDL_MESSAGEBOX_BUTTONS_LEFT_TO_RIGHT = 0x00000080
SDL_MESSAGEBOX_BUTTONS_RIGHT_TO_LEFT = 0x00000100


SDL_MessageBoxButtonFlags = c_int

SDL_MESSAGEBOX_BUTTON_RETURNKEY_DEFAULT = 0x00000001
SDL_MESSAGEBOX_BUTTON_ESCAPEKEY_DEFAULT = 0x00000002


class SDL_MessageBoxButtonData(Structure):
    _fields_ = [("flags", Uint32), ("buttonid", c_int), ("text", c_char_p)]

class SDL_MessageBoxColor(Structure):
    _fields_ = [("r", Uint8), ("g", Uint8), ("b", Uint8)]


SDL_MessageBoxColorType = c_int

SDL_MESSAGEBOX_COLOR_BACKGROUND = 0
SDL_MESSAGEBOX_COLOR_TEXT = 1
SDL_MESSAGEBOX_COLOR_BUTTON_BORDER = 2
SDL_MESSAGEBOX_COLOR_BUTTON_BACKGROUND = 3
SDL_MESSAGEBOX_COLOR_BUTTON_SELECTED = 4
SDL_MESSAGEBOX_COLOR_MAX = 5


class SDL_MessageBoxColorScheme(Structure):
    _fields_ = [("colors", (SDL_MessageBoxColor * SDL_MESSAGEBOX_COLOR_MAX))]

class SDL_MessageBoxData(Structure):
    _fields_ = [("flags", Uint32),
                ("window", POINTER(SDL_Window)),
                ("title", c_char_p),
                ("message", c_char_p),
                ("numbuttons", c_int),
                ("buttons", POINTER(SDL_MessageBoxButtonData)),
                ("colorScheme", POINTER(SDL_MessageBoxColorScheme))
                ]


SDL_ShowMessageBox = _bind("SDL_ShowMessageBox", [POINTER(SDL_MessageBoxData), POINTER(c_int)], c_int)
SDL_ShowSimpleMessageBox = _bind("SDL_ShowSimpleMessageBox", [Uint32, c_char_p, c_char_p, POINTER(SDL_Window)], c_int)



import os
import sys
import pytest
from ctypes import cast, c_char_p, addressof
from sdl2 import filesystem, SDL_free


def test_SDL_GetBasePath():
    path = filesystem.SDL_GetBasePath()
    path = path.decode("utf-8")
    if sys.version_info[0] < 3:
        is_python_path = False
        for s in [u"python", u"pypy", u"pyenv", u"virtualenv"]:
            if s in path.lower():
                is_python_path = True
                break
        assert is_python_path
    else:
        execprefix = sys.base_exec_prefix
        assert execprefix.lower() in path.lower()

def test_SDL_GetPrefPath():
    path = filesystem.SDL_GetPrefPath(b"OrgName", b"AppName")
    path = path.decode("utf-8")
    assert "OrgName" in path
    assert "AppName" in path



from .. import surface, rect, pixels
from .common import SDLError, raise_sdl_err
from .sprite import SoftwareSprite
from .surface import _get_target_surface
from .image import load_bmp

__all__ = ["BitmapFont"]


class BitmapFont(object):
    """A class for rendering text using a given bitmap font.

    This class takes an image of equally-spaced font characters and a font map
    indicating the location of each character on the image, and uses these to
    render text using the given font. This class is based on base SDL2 functions
    and does not require the SDL_ttf library to be installed.

    The font mapping table is a list of strings, with each string representing
    a row of characters on the font image surface. Each character within each
    line is assumed to be of equal height and width, but this can be adjusted
    using the :meth:`remap` method.
    
    For example, the built-in bitmap font ``font.bmp`` has the following layout:

    .. image:: images/font.png

    The default font mapping table, which matches the layout of ``font.bmp``,
    looks like this::

       fontmap = [
           '0123456789',
           'ABCDEFGHIJ',
           'KLMNOPQRST',
           'UVWXYZ    ',
           'abcdefghij',
           'klmnopqrst',
           'uvwxyz    ',
           ',;.:!?+-()',
        ]
    
    Args:
        font_img (:obj:`SDL_Surface` or str): A surface or path to a file
            containing a valid bitmap (``.bmp``) font image.
        size (tuple, optional): A ``(width, height)`` tuple defining the size of
            each character in the bitmap font. If not specified, this will be
            inferred automatically from the fontmap and font image.
        fontmap (list, optional): A list of strings defining the locations of
            characters in the font image. If not specified, the default font map
            defined above will be used.

    """
    DEFAULTMAP = [
        "0123456789",
        "ABCDEFGHIJ",
        "KLMNOPQRST",
        "UVWXYZ    ",
        "abcdefghij",
        "klmnopqrst",
        "uvwxyz    ",
        ",;.:!?+-()",
    ]

    def __init__(self, font_img, size=None, mapping=None):
        if hasattr(font_img, "upper"):  # if string
            self.surface = load_bmp(font_img)
        elif isinstance(font_img, SoftwareSprite):
            self.surface = font_img.surface
            self._sprite = font_img  # prevent GC on the Sprite
        elif isinstance(font_img, surface.SDL_Surface):
            self.surface = font_img
        elif "SDL_Surface" in str(type(font_img)):
            self.surface = font_img.contents
        else:
            raise TypeError("font_img must be a Sprite or SDL_Surface")

        if mapping is None:
            self.mapping = list(BitmapFont.DEFAULTMAP)
        else:
            self.mapping = mapping

        if not size:
            map_rows = len(self.mapping)
            map_cols = len(self.mapping[0])
            surf_size = (float(self.surface.w), float(self.surface.h))
            size = (int(surf_size[0] / map_cols), int(surf_size[1] / map_rows))
        self.size = size[0], size[1]

        self.offsets = {}
        self._max_height = self.size[1]
        self._calculate_offsets()

    def _calculate_offsets(self):
        # Calculates the internal character offsets for each line
        self.offsets = {}
        x, y = 0, 0
        w, h = self.size
        for line in self.mapping:
            x = 0
            for c in line:
                if c not in self.offsets.keys():
                    self.offsets[c] = rect.SDL_Rect(x, y, w, h)
                x += w
            y += h

    def _validate_chars(self, text):
        e = "The character '{0}' does not exist within the current font mapping"
        for ch in text:
            if ch != "\n" and ch not in self.offsets.keys():
                raise ValueError(e.format(ch))

    def _get_rendered_size(self, text, line_h):
        line_h = self._max_height if not line_h else line_h
        text_w, text_h = (0, 0)
        lines = text.split("\n")
        for line in lines:
            line_w = 0
            for c in line:
                charsize = self.offsets[c]
                line_w += charsize.w
            if line_w > text_w:
                text_w = line_w
            text_h += line_h
        return (text_w, text_h)

    def _render_text(self, target, fontsf, lines, line_h, offset=(0, 0)):
        line_h = self._max_height if not line_h else line_h
        dstr = rect.SDL_Rect(0, 0, 0, 0)
        y = offset[1]
        for line in lines:
            x = offset[0]
            for c in line:
                dstr.x = x
                dstr.y = y + (line_h - self.offsets[c].h)
                surface.SDL_BlitSurface(fontsf, self.offsets[c], target, dstr)
                x += self.offsets[c].w
            y += line_h
        return (x, y)

    def remap(self, c, x, y, w, h):
        """Updates the source rectangle for a given font character.

        This method can be used to fine-tune the character mappings in the font
        image to produce better spacing in the rendered text.

        Args:
            c (str): The character to remap in the font image.
            x (int): The x coordinate (in pixels) of the top-left corner of the
                new rectangle for the character.
            y (int): The y coordinate (in pixels) of the top-left corner of the
                new rectangle for the character.
            w (int): The width (in pixels) of the new rectangle for the
                character.
            h (int): The height (in pixels) of the new rectangle for the
                character.

        """
        if len(c) > 1:
            raise ValueError("Can only remap one character at a time.")
        self._validate_chars(c)
        x, y, w, h = [int(i) for i in (x, y, w, h)]
        if any([w < 1, h < 1]):
            raise ValueError("Width and height must both be positive integers.")
        surf_w, surf_h = (self.surface.w, self.surface.h)
        if x < 0 or y < 0 or x+w >= surf_w or y+h >= surf_h:
            e = "Character rectangle cannot exceed the bounds of the font image"
            raise ValueError(e + " ({0}, {1}).".format(surf_w, surf_h))

        self.offsets[c] = rect.SDL_Rect(x, y, w, h)
        if h > self._max_height:
            self._max_height = h

    def render(self, text, bpp=None):
        # Deprecated: replaced by render_text, which returns a surface
        self._validate_chars(text)
        lines = text.split("\n")

        tw, th = self._get_rendered_size(text, None)
        if bpp is None:
            bpp = self.surface.format.contents.BitsPerPixel
        sf = surface.SDL_CreateRGBSurface(0, tw, th, bpp, 0, 0, 0, 0)
        if not sf:
            raise SDLError()
        imgsurface = SoftwareSprite(sf.contents, False)

        self._render_text(imgsurface.surface, self.surface, lines, None)
        return imgsurface

    def render_text(self, text, line_h=None, as_argb=True):
        """Renders a string of text to a new surface.

        If a newline character (``\\n``) is encountered in the string, it will
        be rendered as a line break in the rendered text.

        By default, this function also converts the rendered text from the native
        format of the font image to 32-bit ARGB, for consistency across functions
        and better compatibility with SDL2 renderers. To disable ARGB conversion,
        set the ``as_argb`` parameter to ``False``.

        Args:
            text (str): The string of text to render.
            line_h (int, optional): The line height (in pixels) to use for each
                line of the rendered text. If not specified, the maximum
                character height for the font will be used. Defaults to ``None``.
            as_argb (bool, optional): Whether the output surface should be
                converted to 32-bit ARGB pixel format or left as-is. Defaults to
                ``True`` (convert to ARGB).

        Returns:
            :obj:`~sdl2.SDL_Surface`: A surface containing the rendered text.

        Raises:
            ValueError: If a character in the text is not provided by the
            current font.

        """
        self._validate_chars(text)
        lines = text.split("\n")

        # Create a new surface with the same format as the font image
        tw, th = self._get_rendered_size(text, line_h)
        fmt = self.surface.format.contents.format
        bpp = 32  # according to SDL2 source, this has no effect
        sf = surface.SDL_CreateRGBSurfaceWithFormat(0, tw, th, bpp, fmt)
        if not sf:
            raise_sdl_err("creating the font surface")

        # Render text to the new surface, converting pixel format if necessary
        self._render_text(sf.contents, self.surface, lines, line_h)
        if as_argb and fmt != pixels.SDL_PIXELFORMAT_ARGB8888:
            out_fmt = pixels.SDL_AllocFormat(pixels.SDL_PIXELFORMAT_ARGB8888)
            sf_argb = surface.SDL_ConvertSurface(sf, out_fmt, 0)
            surface.SDL_FreeSurface(sf)
            sf = sf_argb
            if not sf:
                raise_sdl_err("converting rendered text to ARGB format")

        return sf.contents

    def render_on(self, target, text, offset=(0, 0), line_h=None):
        """Renders a string of text to an existing surface.

        If a newline character (``\\n``) is encountered in the string, it will
        be rendered as a line break in the rendered text.

        Args:
            target (:obj:`~sdl2.SDL_Surface`): The surface on which to render
                the given string.
            text (str): The string of text to render to the target surface.
            offset (tuple, optional): The ``(x, y)`` coordinates of the target
                surface on which the top-left corner of the rendered text will
                be placed. Defaults to ``(0, 0)``.
            line_h (int, optional): The line height (in pixels) to use for each
                line of the rendered text. If not specified, the maximum
                character height for the font will be used. Defaults to ``None``.

        Returns:
            tuple: The ``(x1, y1, x2, y2)`` rectangle of the target surface on
            which the text was rendered.

        Raises:
            ValueError: If a character in the text is not provided by the
            current font.

        """
        x1, y1 = offset
        dest = rect.SDL_Rect(x1, y1, 0, 0)
        target = _get_target_surface(target)

        sf = self.render_text(text, line_h, as_argb=False)
        ret = surface.SDL_BlitSurface(sf, None, target, dest)
        if ret != 0:
            raise_sdl_err("copying the text to the target surface")

        return (x1, y1, x1 + sf.w, y1 + sf.h)

    def contains(self, c):
        """Checks whether a given character is mapped within the font.
        
        Args:
            c (str): The character to check for within the font map.

        Returns:
            bool: ``True`` if the font contains the character, otherwise
            ``False``.
    
        """
        return c in self.offsets

    def can_render(self, text):
        # Deprecated: already throws informative exception on missing character
        lines = text.split("\n")
        for line in lines:
            for c in line:
                if c != ' ' and c not in self.offsets:
                    return False
        return True



import sys
import pytest
import sdl2
from sdl2 import (
    SDL_Init, SDL_Quit, SDL_ClearError, SDL_GetError,
    SDL_FALSE, SDL_TRUE, SDL_IGNORE, SDL_ENABLE, SDL_QUERY
)
from sdl2 import joystick
from sdl2 import gamecontroller as gamepad

# Get status of gamepad support/availability before running tests
SDL_ClearError()
ret = SDL_Init(sdl2.SDL_INIT_GAMECONTROLLER)
gamepad_works = ret == 0 and SDL_GetError() == b""
SDL_Quit()

pytestmark = pytest.mark.skipif(not gamepad_works, reason="system unsupported")

gamepad_types = [
    gamepad.SDL_CONTROLLER_TYPE_UNKNOWN,
    gamepad.SDL_CONTROLLER_TYPE_XBOX360,
    gamepad.SDL_CONTROLLER_TYPE_XBOXONE,
    gamepad.SDL_CONTROLLER_TYPE_PS3,
    gamepad.SDL_CONTROLLER_TYPE_PS4,
    gamepad.SDL_CONTROLLER_TYPE_NINTENDO_SWITCH_PRO,
    gamepad.SDL_CONTROLLER_TYPE_VIRTUAL,
    gamepad.SDL_CONTROLLER_TYPE_PS5,
    gamepad.SDL_CONTROLLER_TYPE_AMAZON_LUNA,
    gamepad.SDL_CONTROLLER_TYPE_GOOGLE_STADIA,
]


# Overrides global fixture with one that initializes the joystick system
@pytest.fixture(scope="module")
def with_sdl():
    sdl2.SDL_ClearError()
    sdl2.SDL_SetHint(b"SDL_JOYSTICK_ALLOW_BACKGROUND_EVENTS", b"1")
    ret = sdl2.SDL_Init(sdl2.SDL_INIT_VIDEO | sdl2.SDL_INIT_GAMECONTROLLER)
    assert sdl2.SDL_GetError() == b""
    assert ret == 0
    # Also initialize a virtual joystick (if supported)
    if sdl2.dll.version >= 2014:
        virt_type = joystick.SDL_JOYSTICK_TYPE_GAMECONTROLLER
        virt_index = joystick.SDL_JoystickAttachVirtual(virt_type, 2, 4, 1)
    yield
    sdl2.SDL_Quit()

@pytest.fixture()
def gamepads(with_sdl):
    devices = []
    count = joystick.SDL_NumJoysticks()
    for i in range(count):
        if gamepad.SDL_IsGameController(i) == SDL_TRUE:
            pad = gamepad.SDL_GameControllerOpen(i)
            assert sdl2.SDL_GetError() == b""
            assert isinstance(pad.contents, gamepad.SDL_GameController)
            devices.append(pad)
    yield devices
    for pad in devices:
        gamepad.SDL_GameControllerClose(pad)

def is_virtual(pad):
    stick = gamepad.SDL_GameControllerGetJoystick(pad)
    return joystick.SDL_JoystickName(stick) == b"Virtual Joystick"


# Test if SDL_GameControllerMappingForGUID is able to be tested
if sys.version_info >= (3, 8, 0) or sdl2.dll.version >= 2006:
    has_mapping_for_guid = True
else:
    has_mapping_for_guid = False


def test_SDL_GameControllerAddMapping(with_sdl):
    newmap = (
        b"030000005e0400002700000006010000,Microsoft SideWinder,"
        b"platform:Mac OS X,a:b0,b:b1,x:b2,y:b3,dpup:-a1,dpdown:+a1,"
        b"dpleft:-a0,dpright:+a0,lefttrigger:b4,righttrigger:b5"
    )
    if sdl2.dll.version >= 2006:
        n1 = gamepad.SDL_GameControllerNumMappings()
        ret = gamepad.SDL_GameControllerAddMapping(newmap)
        assert ret >= 0
        n2 = gamepad.SDL_GameControllerNumMappings()
        assert n2 == n1 + 1
    else:
        # NumMappings not available before 2.0.6
        ret = gamepad.SDL_GameControllerAddMapping(newmap)
        assert ret != -1

@pytest.mark.skipif(not has_mapping_for_guid, reason="not available")
def test_SDL_GameControllerMappingForGUID(with_sdl):
    newmap = (
        b"030000005e0400002700000006010000,Microsoft SideWinder,"
        b"platform:Mac OS X,a:b0,b:b1,x:b2,y:b3,dpup:-a1,dpdown:+a1,"
        b"dpleft:-a0,dpright:+a0,lefttrigger:b4,righttrigger:b5"
    )
    ret = gamepad.SDL_GameControllerAddMapping(newmap)
    assert ret >= 0
    # Get GUID for new mapping
    guid_str = newmap.split(b",")[0]
    guid = joystick.SDL_JoystickGetGUIDFromString(guid_str)
    # Get mapping for GUID
    retmap = gamepad.SDL_GameControllerMappingForGUID(guid)
    assert retmap == newmap

def test_SDL_IsGameController(with_sdl):
    count = joystick.SDL_NumJoysticks()
    for index in range(count):
        ret = gamepad.SDL_IsGameController(index)
        assert sdl2.SDL_GetError() == b""
        assert ret in [SDL_TRUE, SDL_FALSE]

def test_SDL_GameControllerNameForIndex(with_sdl):
    count = joystick.SDL_NumJoysticks()
    for index in range(count):
        name = gamepad.SDL_GameControllerNameForIndex(index)
        assert sdl2.SDL_GetError() == b""
        assert name == None or type(name) in (str, bytes)

@pytest.mark.skipif(sdl2.dll.version < 2012, reason="not available")
def test_SDL_GameControllerTypeForIndex(with_sdl):
    count = joystick.SDL_NumJoysticks()
    for index in range(count):
        padtype = gamepad.SDL_GameControllerTypeForIndex(index)
        assert padtype in gamepad_types

def test_SDL_GameControllerOpenClose(with_sdl):
    count = joystick.SDL_NumJoysticks()
    for index in range(count):
        if gamepad.SDL_IsGameController(index) == SDL_TRUE:
            pad = gamepad.SDL_GameControllerOpen(index)
            assert sdl2.SDL_GetError() == b""
            assert isinstance(pad.contents, gamepad.SDL_GameController)
            gamepad.SDL_GameControllerClose(pad)

def test_SDL_GameControllerMapping(gamepads):
    for pad in gamepads:
        mapping = gamepad.SDL_GameControllerMapping(pad)
        assert SDL_GetError() == b""
        assert mapping == None or type(mapping) in (str, bytes)

def test_SDL_GameControllerName(gamepads):
    names = []
    for pad in gamepads:
        name = gamepad.SDL_GameControllerName(pad)
        assert type(name) in (str, bytes)
        names.append(name.decode('utf-8'))
    print(names)

@pytest.mark.skipif(sdl2.dll.version < 2012, reason="not available")
def test_SDL_GameControllerGetType(gamepads):
    for pad in gamepads:
        padtype = gamepad.SDL_GameControllerGetType(pad)
        assert SDL_GetError() == b""
        assert padtype in gamepad_types
        if is_virtual(pad):
            assert padtype == gamepad.SDL_CONTROLLER_TYPE_VIRTUAL

def test_SDL_GameControllerGetAttached(gamepads):
    for pad in gamepads:
        ret = gamepad.SDL_GameControllerGetAttached(pad)
        assert ret in [SDL_FALSE, SDL_TRUE]

def test_SDL_GameControllerGetJoystick(gamepads):
    for pad in gamepads:
        stick = gamepad.SDL_GameControllerGetJoystick(pad)
        assert SDL_GetError() == b""
        assert isinstance(stick.contents, joystick.SDL_Joystick)

def test_SDL_GameControllerEventState(with_sdl):
    for state in (SDL_IGNORE, SDL_ENABLE):
        news = gamepad.SDL_GameControllerEventState(state)
        assert news == state
        query = gamepad.SDL_GameControllerEventState(SDL_QUERY)
        assert query == state

def test_SDL_GameControllerUpdate(with_sdl):
    # NOTE: Returns void, can't really test anything else
    gamepad.SDL_GameControllerUpdate()
    assert SDL_GetError() == b""

def test_SDL_GameControllerGetAxisFromString(with_sdl):
    expected = {
        b'lefty': gamepad.SDL_CONTROLLER_AXIS_LEFTY,
        b'lefttrigger': gamepad.SDL_CONTROLLER_AXIS_TRIGGERLEFT,
        b'notanaxis': gamepad.SDL_CONTROLLER_AXIS_INVALID
    }
    for string in expected.keys():
        a = gamepad.SDL_GameControllerGetAxisFromString(string)
        assert a == expected[string]

def test_SDL_GameControllerGetStringForAxis(with_sdl):
    expected = {
        gamepad.SDL_CONTROLLER_AXIS_LEFTY: b'lefty',
        gamepad.SDL_CONTROLLER_AXIS_TRIGGERLEFT: b'lefttrigger',
        gamepad.SDL_CONTROLLER_AXIS_INVALID: None
    }
    for axis in expected.keys():
        s = gamepad.SDL_GameControllerGetStringForAxis(axis)
        assert s == expected[axis]

@pytest.mark.skip("not implemented")
def test_SDL_GameControllerGetBindForAxis(gamepads):
    pass

@pytest.mark.skip("not implemented")
@pytest.mark.skipif(sdl2.dll.version < 2014, reason="not available")
def test_SDL_GameControllerHasAxis(gamepads):
    pass

@pytest.mark.skip("not implemented")
def test_SDL_GameControllerGetAxis(gamepads):
    pass

def test_SDL_GameControllerGetButtonFromString(with_sdl):
    expected = {
        b'x': gamepad.SDL_CONTROLLER_BUTTON_X,
        b'dpup': gamepad.SDL_CONTROLLER_BUTTON_DPAD_UP,
        b'notabutton': gamepad.SDL_CONTROLLER_BUTTON_INVALID
    }
    for string in expected.keys():
        b = gamepad.SDL_GameControllerGetButtonFromString(string)
        assert b == expected[string]

def test_SDL_GameControllerGetStringForButton(with_sdl):
    expected = {
        gamepad.SDL_CONTROLLER_BUTTON_X: b'x',
        gamepad.SDL_CONTROLLER_BUTTON_DPAD_UP: b'dpup',
        gamepad.SDL_CONTROLLER_BUTTON_INVALID: None
    }
    for button in expected.keys():
        s = gamepad.SDL_GameControllerGetStringForButton(button)
        assert s == expected[button]

@pytest.mark.skip("not implemented")
def test_SDL_GameControllerGetBindForButton(gamepads):
    pass

@pytest.mark.skip("not implemented")
@pytest.mark.skipif(sdl2.dll.version < 2014, reason="not available")
def test_SDL_GameControllerHasButton(gamepads):
    pass

@pytest.mark.skip("not implemented")
def test_SDL_GameControllerGetButton(gamepads):
    pass

@pytest.mark.skip("not implemented")
@pytest.mark.skipif(sdl2.dll.version < 2014, reason="not available")
def test_SDL_GameControllerGetNumTouchpads(gamepads):
    pass

@pytest.mark.skip("not implemented")
@pytest.mark.skipif(sdl2.dll.version < 2014, reason="not available")
def test_SDL_GameControllerGetNumTouchpadFingers(gamepads):
    pass

@pytest.mark.skip("not implemented")
@pytest.mark.skipif(sdl2.dll.version < 2014, reason="not available")
def test_SDL_GameControllerGetTouchpadFinger(gamepads):
    pass

@pytest.mark.skip("not implemented")
@pytest.mark.skipif(sdl2.dll.version < 2014, reason="not available")
def test_SDL_GameControllerHasSensor(gamepads):
    pass

@pytest.mark.skip("not implemented")
@pytest.mark.skipif(sdl2.dll.version < 2014, reason="not available")
def test_SDL_GameControllerSetSensorEnabled(gamepads):
    pass

@pytest.mark.skip("not implemented")
@pytest.mark.skipif(sdl2.dll.version < 2014, reason="not available")
def test_SDL_GameControllerIsSensorEnabled(gamepads):
    pass

@pytest.mark.skip("not implemented")
@pytest.mark.skipif(sdl2.dll.version < 2016, reason="not available")
def test_SDL_GameControllerGetSensorDataRate(gamepads):
    pass

@pytest.mark.skip("not implemented")
@pytest.mark.skipif(sdl2.dll.version < 2014, reason="not available")
def test_SDL_GameControllerGetSensorData(gamepads):
    pass

@pytest.mark.skip("not implemented")
def test_SDL_GameControllerAddMappingsFromRW(gamepads):
    pass

@pytest.mark.skip("not implemented")
def test_SDL_GameControllerAddMappingsFromFile(gamepads):
    pass

def test_SDL_GameControllerFromInstanceID(gamepads):
    for pad in gamepads:
        stick = gamepad.SDL_GameControllerGetJoystick(pad)
        iid = joystick.SDL_JoystickInstanceID(stick)
        assert iid >= 0
        pad2 = gamepad.SDL_GameControllerFromInstanceID(iid)
        name = gamepad.SDL_GameControllerName(pad)
        assert gamepad.SDL_GameControllerName(pad2) == name

@pytest.mark.skipif(sdl2.dll.version < 2012, reason="not available")
def test_SDL_GameControllerFromPlayerIndex(gamepads):
    i = 0
    for pad in gamepads:
        gamepad.SDL_GameControllerSetPlayerIndex(pad, i)
        pad2 = gamepad.SDL_GameControllerFromPlayerIndex(i)
        name = gamepad.SDL_GameControllerName(pad)
        assert gamepad.SDL_GameControllerName(pad2) == name
        i += 1

@pytest.mark.skipif(sdl2.dll.version < 2009, reason="not available")
def test_SDL_GameControllerGetPlayerIndex(gamepads):
    for pad in gamepads:
        player = gamepad.SDL_GameControllerGetPlayerIndex(pad)
        assert player in [-1, 0, 1, 2, 3]

@pytest.mark.skipif(sdl2.dll.version < 2012, reason="not available")
def test_SDL_GameControllerSetPlayerIndex(gamepads):
    i = 0
    for pad in gamepads:
        gamepad.SDL_GameControllerSetPlayerIndex(pad, i)
        player = gamepad.SDL_GameControllerGetPlayerIndex(pad)
        assert player == i
        i += 1

@pytest.mark.skipif(sdl2.dll.version < 2006, reason="not available")
def test_SDL_GameControllerGetVendor(gamepads):
    for pad in gamepads:
        vid = gamepad.SDL_GameControllerGetVendor(pad)
        assert SDL_GetError() == b""
        if not is_virtual(pad):
            assert vid > 0

@pytest.mark.skipif(sdl2.dll.version < 2006, reason="not available")
def test_SDL_GameControllerGetProduct(gamepads):
    for pad in gamepads:
        pid = gamepad.SDL_GameControllerGetProduct(pad)
        assert SDL_GetError() == b""
        if not is_virtual(pad):
            assert pid > 0

@pytest.mark.skipif(sdl2.dll.version < 2006, reason="not available")
def test_SDL_GameControllerGetProductVersion(gamepads):
    for pad in gamepads:
        pver = gamepad.SDL_GameControllerGetProductVersion(pad)
        assert SDL_GetError() == b""
        assert pver >= 0

@pytest.mark.skipif(sdl2.dll.version < 2014, reason="not available")
def test_SDL_GameControllerGetSerial(gamepads):
    for pad in gamepads:
        serial = gamepad.SDL_GameControllerGetSerial(pad)
        assert SDL_GetError() == b""
        assert serial == None or type(serial) in (str, bytes)

@pytest.mark.skipif(sdl2.dll.version < 2006, reason="not available")
def test_SDL_GameControllerNumMappings(with_sdl):
    num = gamepad.SDL_GameControllerNumMappings()
    assert num > 0

@pytest.mark.skipif(sdl2.dll.version < 2006, reason="not available")
def test_SDL_GameControllerMappingForIndex(with_sdl):
    newmap = (
        b"030000005e0400002700000006010000,Microsoft SideWinder,"
        b"platform:Mac OS X,a:b0,b:b1,x:b2,y:b3,dpup:-a1,dpdown:+a1,"
        b"dpleft:-a0,dpright:+a0,lefttrigger:b4,righttrigger:b5"
    )
    ret = gamepad.SDL_GameControllerAddMapping(newmap)
    assert ret >= 0
    num = gamepad.SDL_GameControllerNumMappings()
    retmap = gamepad.SDL_GameControllerMappingForIndex(num - 1)
    assert newmap == retmap

@pytest.mark.skip("not implemented")
@pytest.mark.skipif(sdl2.dll.version < 2009, reason="not available")
def test_SDL_GameControllerMappingForDeviceIndex(with_sdl):
    pass

@pytest.mark.skipif(sdl2.dll.version < 2009, reason="not available")
def test_SDL_GameControllerRumble(gamepads):
    # If we ever add an interactive test suite, this should be moved there
    for pad in gamepads:
        # 50% strength low-frequency, 25% high-frequency rumble for 500ms
        ret = gamepad.SDL_GameControllerRumble(pad, 32767, 16384, 500)
        assert ret in [-1, 0]

@pytest.mark.skipif(sdl2.dll.version < 2014, reason="not available")
def test_SDL_GameControllerRumbleTriggers(gamepads):
    # If we ever add an interactive test suite, this should be moved there
    for pad in gamepads:
        # 50% strength left trigger, 25% right trigger rumble for 500ms
        ret = gamepad.SDL_GameControllerRumbleTriggers(pad, 32767, 16384, 500)
        assert ret in [-1, 0]

@pytest.mark.skipif(sdl2.dll.version < 2014, reason="not available")
def test_SDL_GameControllerHasSetLED(gamepads):
    # If we ever add an interactive test suite, this should be moved there
    for pad in gamepads:
        has_led = gamepad.SDL_GameControllerHasLED(pad)
        assert has_led in [SDL_FALSE, SDL_TRUE]
        expected = -1 if has_led == SDL_FALSE else 0
        cols = [(255, 0, 0), (0, 255, 0), (0, 0, 255)]
        for r, g, b in cols:
            ret = gamepad.SDL_GameControllerSetLED(pad, r, g, b)
            assert ret == expected

@pytest.mark.skipif(sdl2.dll.version < 2018, reason="not available")
def test_SDL_GameControllerHasRumble(gamepads):
    # If we ever add an interactive test suite, this should be moved there
    for pad in gamepads:
        has_rumble = gamepad.SDL_GameControllerHasRumble(pad)
        assert has_rumble in [SDL_FALSE, SDL_TRUE]

@pytest.mark.skipif(sdl2.dll.version < 2018, reason="not available")
def test_SDL_GameControllerHasRumbleTriggers(gamepads):
    # If we ever add an interactive test suite, this should be moved there
    for pad in gamepads:
        has_rumble_triggers = gamepad.SDL_GameControllerHasRumbleTriggers(pad)
        assert has_rumble_triggers in [SDL_FALSE, SDL_TRUE]

@pytest.mark.skip("not implemented")
@pytest.mark.skipif(sdl2.dll.version < 2016, reason="not available")
def test_SDL_GameControllerSendEffect(gamepads):
    # Probably impossible to test since effect data would be specific
    # to each controller type?
    pass

@pytest.mark.skip("Only relevant on iOS, not testable by PySDL2")
@pytest.mark.skipif(sdl2.dll.version < 2018, reason="not available")
def test_SDL_GameControllerGetAppleSFSymbolsNameForButtonAxis(gamepads):
    # The following two functions are deliberatly ignored:
    # - SDL_GameControllerGetAppleSFSymbolsNameForButton
    # - SDL_GameControllerGetAppleSFSymbolsNameForAxis
    pass



import sys
import os
import re
import zipfile
import tarfile
import io

__all__ = ["open_zipfile", "open_tarfile", "open_url", "Resources"]

# Python 3.x workarounds for the changed urllib modules.
if sys.version_info[0] >= 3:
    import urllib.parse as urlparse
    import urllib.request as urllib2
else:
    import urlparse
    import urllib2


def _validate_path(path, what, write=False):
    fullpath = os.path.abspath(path)
    fname = os.path.basename(path)
    if write:
        parent = os.path.abspath(os.path.join(fullpath, os.pardir))
        if not os.path.isdir(parent):
            e = "The given parent directory '{0}' does not exist"
            raise IOError(e.format(parent))
    else:
        if not os.path.exists(fullpath):
            e = "Could not find {0} at the given path: {1}"
            raise IOError(e.format(what, fullpath))
    return (fullpath, fname)


def open_zipfile(archive, filename, directory=None):
    """Retrieves a given file from a ZIP archive.

    Args:
        archive (:obj:`~zipfile.ZipFile`, str): The ZipFile object or path to
            the ZIP archive containing the desired file.
        filename (str): The name of the file to retrieve from the archive.
        directory (str, optional): The path to the directory within the archive
            containing the file to retrieve. Defaults to the root level of the
            archive.

    Returns:
        :obj:`~io.BytesIO`: A Python bytestream object containing the requested
        file.

    Raises:
        KeyError: If the given file could not be found within the archive.
        TypeError: If the archive is not a valid ZIP archive.

    """
    data = None
    opened = False

    if not isinstance(archive, zipfile.ZipFile):
        if not zipfile.is_zipfile(archive):
            raise TypeError("passed file does not seem to be a ZIP archive")
        else:
            archive = zipfile.ZipFile(archive, 'r')
            opened = True

    apath = filename
    if directory:
        apath = "%s/%s" % (directory, filename)

    try:
        dmpdata = archive.open(apath)
        data = io.BytesIO(dmpdata.read())
    finally:
        if opened:
            archive.close()
    return data


def open_tarfile(archive, filename, directory=None, ftype=None):
    """Retrieves a given file from a TAR archive.

    If the TAR archive uses ``.tar.gz`` or ``.tar.bz2`` compression and the
    file name does not contain either of these identifiers, the compression
    type must be manually specified.

    Args:
        archive (:obj:`~tarfile.TarFile`, str): The TarFile object or path to
            the TAR archive containing the desired file.
        filename (str): The name of the file to retrieve from the archive.
        directory (str, optional): The path to the directory within the archive
            containing the file to retrieve. Defaults to the root level of the
            archive.
        ftype (str, optional): The compression type (if any) used for the TAR
            file, can be either 'gz', 'bz2', or None (no compression). If not
            specified, will default to assuming no compression.

    Returns:
        :obj:`~io.BytesIO`: A Python bytestream object containing the requested
        file.

    Raises:
        KeyError: If the given file could not be found within the archive.
        TypeError: If the archive is not a supported TAR archive.

    """
    data = None
    opened = False

    if not isinstance(archive, tarfile.TarFile):
        if not tarfile.is_tarfile(archive):
            raise TypeError("passed file does not seem to be a TAR archive")
        else:
            file_ext = archive.split('.')[-1]
            if not ftype and file_ext in ('gz', 'bz2'):
                ftype = file_ext
            if ftype and ftype not in ('gz', 'bz2'):
                e = "invalid TAR compression type '{0}' (must be 'gz' or 'bz2')"
                raise TypeError(e.format(ftype))
            mode = 'r:{0}'.format(ftype) if ftype else 'r'
            archive = tarfile.open(archive, mode)
            opened = True

    apath = filename
    if directory:
        apath = "%s/%s" % (directory, filename)

    try:
        dmpdata = archive.extractfile(apath)
        data = io.BytesIO(dmpdata.read())
    finally:
        if opened:
            archive.close()
    return data


def open_url(filename, basepath=None):
    # Opens and reads a certain file from a web or remote location.
    # Deprecated because its argument names are confusing and because
    # users are probably better off using urllib directly.
    url = filename
    if basepath:
        url = urlparse.urljoin(basepath, filename)
    return urllib2.urlopen(url)


class Resources(object):
    """A container class for managing application resource files.
    
    This class eases access to resources by allowing access using relative
    paths, scanning archives to locate files, and more.

    Args:
        path (str, optional): The path of a resource directory with which to
            initialze the container. Defaults to ``None``.
        subdir (str, optional): Deprecated, do not use.
        excludepattern (str, optional): A regular expression indicating
            which directories (if any) to ignore if initializing the
            container with a resource path. Defaults to ``None``.

    """
    def __init__(self, path=None, subdir=None, excludepattern=None):
        self.files = {}
        if path:
            self.scan(path, subdir, excludepattern)

    def _scanzip(self, filename):
        """Scans the passed ZIP archive and indexes all the files
        contained by it.
        """
        if not zipfile.is_zipfile(filename):
            raise TypeError("file '%s' is not a valid ZIP archive" % filename)
        archname = os.path.abspath(filename)
        zipf = zipfile.ZipFile(filename, 'r')
        for path in zipf.namelist():
            fname = os.path.split(path)[1]
            if fname:
                self.files[fname] = (archname, 'zip', path)
        zipf.close()

    def _scantar(self, filename, ftype=None):
        """Scans the passed TAR archive and indexes all the files
        contained by it.
        """
        if not tarfile.is_tarfile(filename):
            raise TypeError("file '%s' is not a valid TAR archive" % filename)
        file_ext = filename.split('.')[-1]
        if not ftype and file_ext in ('gz', 'bz2'):
            ftype = file_ext
        if ftype and ftype not in ('gz', 'bz2'):
            e = "invalid TAR compression type '{0}' (must be 'gz' or 'bz2')"
            raise TypeError(e.format(ftype))
        mode = 'r:{0}'.format(ftype) if ftype else 'r'
        archname = os.path.abspath(filename)
        archtype = 'tar'
        if ftype:
            archtype = 'tar%s' % ftype
        tar = tarfile.open(filename, mode)
        for path in tar.getnames():
            fname = os.path.split(path)[1]
            self.files[fname] = (archname, archtype, path)
        tar.close()

    def add(self, filename):
        """Adds a file to the Resources container.

        If the given file is a supported archive, its contents will be scanned
        and added to the container.

        Args:
            filename (str): The filepath of the resource to add to the
                container.

        Raises:
            ValueError: If the file does not exist at the provided path.

        """
        if not os.path.exists(filename):
            raise ValueError("invalid file path")
        if zipfile.is_zipfile(filename):
            self.add_archive(filename)
        elif tarfile.is_tarfile(filename):
            self.add_archive(filename, 'tar')
        else:
            self.add_file(filename)

    def add_file(self, filename):
        """Adds a file without scanning to the Resources container.

        Unlike :meth:`add`, this method will not attempt to add the contents
        of any provided archives to the container.

        Args:
            filename (str): The filepath of the resource to add to the
                container.

        Raises:
            ValueError: If the file does not exist at the provided path.

        """
        if not os.path.exists(filename):
            raise ValueError("invalid file path")
        abspath = os.path.abspath(filename)
        fname = os.path.split(abspath)[1]
        if not fname:
            raise ValueError("invalid file path")
        self.files[fname] = (None, None, abspath)

    def add_archive(self, filename, typehint='zip'):
        """Adds a ``.zip`` or ``.tar`` archive to the container.

        This will scan the passed archive and add its contents to the
        list of available resources. Currently ``.zip``, ``.tar``,
        ``.tar.bz2``, and ``.tar.gz`` formats are supported.

        Args:
            filename (str): The filepath of the archive to scan and add to the
                container.
            typehint (str, optional): The format of the archive to add to the
                container, required if using a custom file extension. Must be
                one of ``zip``, ``tar``, ``tarbz2``, or ``targz``. Defaults to
                ``zip`` if not specified.

        Raises:
            ValueError: If the file does not exist at the provided path, or if
                the file is not a supported archive type.

        """
        if not os.path.exists(filename):
            raise ValueError("invalid file path")
        fname = os.path.basename(filename)
        if 'zip' in fname.split('.'):
            self._scanzip(filename)
        elif 'tar' in fname.split('.'):
            self._scantar(filename)
        else:
            if typehint == 'zip':
                self._scanzip(filename)
            elif typehint == 'tar':
                self._scantar(filename)
            elif typehint == 'tarbz2':
                self._scantar(filename, 'bz2')
            elif typehint == 'targz':
                self._scantar(filename, 'gz')
            else:
                raise ValueError("unsupported archive type")

    def get(self, filename):
        """Retrieves a resource file by name from the container.

        Args:
            filename (str): The file name of the resource to retrieve.

        Returns:
            :obj:`~io.BytesIO`: A Python bytestream object containing the
            retrieved resource file.

        Raises:
            KeyError: If the given file could not be found.

        """
        archive, ftype, pathname = self.files[filename]
        if archive:
            if ftype == 'zip':
                return open_zipfile(archive, pathname)
            elif ftype == 'tar':
                return open_tarfile(archive, pathname)
            elif ftype == 'tarbz2':
                return open_tarfile(archive, pathname, ftype='bz2')
            elif ftype == 'targz':
                return open_tarfile(archive, pathname, ftype='gz')
            else:
                raise ValueError("unsupported archive type")
        dmpdata = open(pathname, 'rb')
        data = io.BytesIO(dmpdata.read())
        dmpdata.close()
        return data

    def get_filelike(self, filename):
        # Deprecated, doesn't make much difference in Python 3
        archive, ftype, pathname = self.files[filename]
        if archive:
            if ftype == 'zip':
                return open_zipfile(archive, pathname)
            elif ftype == 'tar':
                return open_tarfile(archive, pathname)
            elif ftype == 'tarbz2':
                return open_tarfile(archive, pathname, ftype='bz2')
            elif ftype == 'targz':
                return open_tarfile(archive, pathname, ftype='gz')
            else:
                raise ValueError("unsupported archive type")
        return open(pathname, 'rb')

    def get_path(self, filename):
        """Gets the path of a given resource file.

        If the file is only available within an archive, a string in the form
        ``filename@archivename`` will be returned.

        Args:
            filename (str): The file name of the resource to locate.

        Returns:
            str: The absolute path of the resource file, or the archive
            identifier string if the resource is inside an archive.

        Raises:
            KeyError: If the given file could not be found.

        """
        archive, ftype, pathname = self.files[filename]
        if archive:
            return '%s@%s' % (pathname, archive)
        return pathname

    def scan(self, path, subdir=None, excludepattern=None):
        """Scans a path, adding all matching files to the container.

        If a located file is a ``.zip`` or ``.tar`` archive, its
        contents will be indexed and added to the container automatically.

        Args:
            path (str): The path of the directory to scan.
            subdir (str, optional): Deprecated, do not use.
            excludepattern (str, optional): A regular expression indicating
                which directories (if any) within the file structure of the
                given path to exclude from indexing. Defaults to ``None``. 

        Raises:
            ValueError: If the specified path does not exist.

        """
        match = None
        if excludepattern:
            match = re.compile(excludepattern).match
        join = os.path.join
        add = self.add
        abspath = os.path.abspath(path)
        if not os.path.exists(abspath):
            raise ValueError("invalid path '%s'" % abspath)
        if not os.path.isdir(abspath):
            abspath = os.path.dirname(abspath)
        if subdir is not None:
            abspath = os.path.join(abspath, subdir)
        if not os.path.exists(abspath):
            raise ValueError("invalid path '%s'" % abspath)
        for (pdir, dirnames, filenames) in os.walk(abspath):
            if match and match(pdir) is not None:
                continue
            for fname in filenames:
                add(join(pdir, fname))



import os
from ctypes import c_int, c_uint, c_long, c_char_p, c_void_p
from ctypes import POINTER as _P
from .dll import DLL, SDLFunc
from .version import SDL_version, SDL_VERSIONNUM
from .rwops import SDL_RWops
from .stdinc import Uint16, Uint32, SDL_bool
from .pixels import SDL_Color
from .surface import SDL_Surface
from .error import SDL_GetError, SDL_SetError

__all__ = [
    # Opaque Types
    "TTF_Font",

    # Enums
    "hb_direction_t",
    "HB_DIRECTION_INVALID", "HB_DIRECTION_LTR", "HB_DIRECTION_RTL",
    "HB_DIRECTION_TTB", "HB_DIRECTION_BTT",

    # Defines
    "SDL_TTF_MAJOR_VERSION", "SDL_TTF_MINOR_VERSION", "SDL_TTF_PATCHLEVEL",
    "TTF_MAJOR_VERSION", "TTF_MINOR_VERSION", "TTF_PATCHLEVEL",
    "UNICODE_BOM_NATIVE", "UNICODE_BOM_SWAPPED",
    "TTF_STYLE_NORMAL", "TTF_STYLE_BOLD", "TTF_STYLE_ITALIC",
    "TTF_STYLE_UNDERLINE", "TTF_STYLE_STRIKETHROUGH",
    "TTF_HINTING_NORMAL", "TTF_HINTING_LIGHT", "TTF_HINTING_MONO",
    "TTF_HINTING_NONE", "TTF_HINTING_LIGHT_SUBPIXEL",

    # Macro Functions
    "SDL_TTF_VERSION",  "TTF_VERSION", "SDL_TTF_COMPILEDVERSION",
    "SDL_TTF_VERSION_ATLEAST", "HB_TAG",

    # Functions not contained in _funcdefs
    "TTF_RenderText", "TTF_RenderUTF8", "TTF_RenderUNICODE",
    "TTF_SetError", "TTF_GetError",

    # Python Functions
    "get_dll_file",
]


try:
    dll = DLL("SDL2_ttf", ["SDL2_ttf", "SDL2_ttf-2.0"],
              os.getenv("PYSDL2_DLL_PATH"))
except RuntimeError as exc:
    raise ImportError(exc)

def get_dll_file():
    """Gets the file name of the loaded SDL2_ttf library."""
    return dll.libfile

_bind = dll.bind_function


# Constants, enums, type definitions, and macros

SDL_TTF_MAJOR_VERSION = 2
SDL_TTF_MINOR_VERSION = 0
SDL_TTF_PATCHLEVEL = 18

def SDL_TTF_VERSION(x):
    x.major = SDL_TTF_MAJOR_VERSION
    x.minor = SDL_TTF_MINOR_VERSION
    x.patch = SDL_TTF_PATCHLEVEL

TTF_MAJOR_VERSION = SDL_TTF_MAJOR_VERSION
TTF_MINOR_VERSION = SDL_TTF_MINOR_VERSION
TTF_PATCHLEVEL = SDL_TTF_PATCHLEVEL
TTF_VERSION = SDL_TTF_VERSION

SDL_TTF_COMPILEDVERSION = SDL_VERSIONNUM(SDL_TTF_MAJOR_VERSION, SDL_TTF_MINOR_VERSION, SDL_TTF_PATCHLEVEL)
SDL_TTF_VERSION_ATLEAST = lambda x, y, z: (SDL_TTF_COMPILEDVERSION >= SDL_VERSIONNUM(x, y, z))

UNICODE_BOM_NATIVE = 0xFEFF
UNICODE_BOM_SWAPPED = 0xFFFE

TTF_STYLE_NORMAL = 0x00
TTF_STYLE_BOLD = 0x01
TTF_STYLE_ITALIC = 0x02
TTF_STYLE_UNDERLINE = 0x04
TTF_STYLE_STRIKETHROUGH = 0x08

TTF_HINTING_NORMAL = 0
TTF_HINTING_LIGHT = 1
TTF_HINTING_MONO = 2
TTF_HINTING_NONE = 3
TTF_HINTING_LIGHT_SUBPIXEL = 4

class TTF_Font(c_void_p):
    """The opaque data type for fonts opened using the TTF library.
    
    This contains all data associated with a loaded font. Once you are done
    with a :obj:`TTF_Font`, it should be freed using :func:`TTF_CloseFont`.

    """
    pass


# Some additional definitions from HarfBuzz for SetDirection/SetScript

hb_direction_t = c_int
HB_DIRECTION_INVALID = 0
HB_DIRECTION_LTR = 4
HB_DIRECTION_RTL = 5
HB_DIRECTION_TTB = 6
HB_DIRECTION_BTT = 7

def HB_TAG(c1, c2, c3, c4):
    """Converts a 4-character ISO 15924 code into a HarfBuzz script constant.

    A full list of possible 4-character script codes can be found
    here: https://unicode.org/iso15924/iso15924-codes.html

    Args:
        c1 (str): The first character of the code.
        c2 (str): The second character of the code.
        c3 (str): The third character of the code.
        c4 (str): The fourth character of the code.

    Returns:
        int: The HarfBuzz contstant corresponding to the given script.

    """
    c1, c2, c3, c4 = [ord(c) & 0xFF for c in (c1, c2, c3, c4)]
    return (c1 << 24 | c2 << 16 | c3 << 8 | c4)


# Raw ctypes function definitions

_funcdefs = [
    SDLFunc("TTF_Linked_Version", None, _P(SDL_version)),
    SDLFunc("TTF_GetFreeTypeVersion", [_P(c_int), _P(c_int), _P(c_int)], added='2.0.18'),
    SDLFunc("TTF_GetHarfBuzzVersion", [_P(c_int), _P(c_int), _P(c_int)], added='2.0.18'),
    SDLFunc("TTF_ByteSwappedUNICODE", [c_int], None),
    SDLFunc("TTF_Init", None, c_int),
    SDLFunc("TTF_OpenFont", [c_char_p, c_int], _P(TTF_Font)),
    SDLFunc("TTF_OpenFontIndex", [c_char_p, c_int, c_long], _P(TTF_Font)),
    SDLFunc("TTF_OpenFontRW", [_P(SDL_RWops), c_int, c_int], _P(TTF_Font)),
    SDLFunc("TTF_OpenFontIndexRW", [_P(SDL_RWops), c_int, c_int, c_long], _P(TTF_Font)),
    SDLFunc("TTF_OpenFontDPI", [c_char_p, c_int, c_uint, c_uint], _P(TTF_Font), added='2.0.18'),
    SDLFunc("TTF_OpenFontIndexDPI", [c_char_p, c_int, c_long, c_uint, c_uint], _P(TTF_Font), added='2.0.18'),
    SDLFunc("TTF_OpenFontDPIRW", [_P(SDL_RWops), c_int, c_int, c_uint, c_uint], _P(TTF_Font), added='2.0.18'),
    SDLFunc("TTF_OpenFontIndexDPIRW", [_P(SDL_RWops), c_int, c_int, c_long, c_uint, c_uint], _P(TTF_Font), added='2.0.18'),
    SDLFunc("TTF_SetFontSize", [_P(TTF_Font), c_int], c_int, added='2.0.18'),
    SDLFunc("TTF_SetFontSizeDPI", [_P(TTF_Font), c_int, c_uint, c_uint], c_int, added='2.0.18'),
    SDLFunc("TTF_GetFontStyle", [_P(TTF_Font)], c_int),
    SDLFunc("TTF_SetFontStyle", [_P(TTF_Font), c_int], None),
    SDLFunc("TTF_GetFontOutline", [_P(TTF_Font)], c_int),
    SDLFunc("TTF_SetFontOutline", [_P(TTF_Font), c_int], None),
    SDLFunc("TTF_GetFontHinting", [_P(TTF_Font)], c_int),
    SDLFunc("TTF_SetFontHinting", [_P(TTF_Font), c_int], None),
    SDLFunc("TTF_FontHeight", [_P(TTF_Font)], c_int),
    SDLFunc("TTF_FontAscent", [_P(TTF_Font)], c_int),
    SDLFunc("TTF_FontDescent", [_P(TTF_Font)], c_int),
    SDLFunc("TTF_FontLineSkip", [_P(TTF_Font)], c_int),
    SDLFunc("TTF_GetFontKerning", [_P(TTF_Font)], c_int),
    SDLFunc("TTF_SetFontKerning", [_P(TTF_Font), c_int]),
    SDLFunc("TTF_FontFaces", [_P(TTF_Font)], c_long),
    SDLFunc("TTF_FontFaceIsFixedWidth", [_P(TTF_Font)], c_int),
    SDLFunc("TTF_FontFaceFamilyName", [_P(TTF_Font)], c_char_p),
    SDLFunc("TTF_FontFaceStyleName", [_P(TTF_Font)], c_char_p),
    SDLFunc("TTF_GlyphIsProvided", [_P(TTF_Font), Uint16], c_int),
    SDLFunc("TTF_GlyphIsProvided32", [_P(TTF_Font), Uint32], c_int, added='2.0.18'),
    SDLFunc("TTF_GlyphMetrics", [_P(TTF_Font), Uint16, _P(c_int), _P(c_int), _P(c_int), _P(c_int), _P(c_int)], c_int),
    SDLFunc("TTF_GlyphMetrics32", [_P(TTF_Font), Uint32, _P(c_int), _P(c_int), _P(c_int), _P(c_int), _P(c_int)], c_int, added='2.0.18'),
    SDLFunc("TTF_SizeText", [_P(TTF_Font), c_char_p, _P(c_int), _P(c_int)], c_int),
    SDLFunc("TTF_SizeUTF8", [_P(TTF_Font), c_char_p, _P(c_int), _P(c_int)], c_int),
    SDLFunc("TTF_SizeUNICODE", [_P(TTF_Font), _P(Uint16), _P(c_int), _P(c_int)], c_int),
    SDLFunc("TTF_MeasureText", [_P(TTF_Font), c_char_p, c_int, _P(c_int), _P(c_int)], c_int, added='2.0.18'),
    SDLFunc("TTF_MeasureUTF8", [_P(TTF_Font), c_char_p, c_int, _P(c_int), _P(c_int)], c_int, added='2.0.18'),
    SDLFunc("TTF_MeasureUNICODE", [_P(TTF_Font), _P(Uint16), c_int, _P(c_int), _P(c_int)], c_int, added='2.0.18'),
    SDLFunc("TTF_RenderText_Solid", [_P(TTF_Font), c_char_p, SDL_Color], _P(SDL_Surface)),
    SDLFunc("TTF_RenderUTF8_Solid", [_P(TTF_Font), c_char_p, SDL_Color], _P(SDL_Surface)),
    SDLFunc("TTF_RenderUNICODE_Solid", [_P(TTF_Font), _P(Uint16), SDL_Color], _P(SDL_Surface)),
    SDLFunc("TTF_RenderText_Solid_Wrapped", [_P(TTF_Font), c_char_p, SDL_Color, Uint32], _P(SDL_Surface), added='2.0.18'),
    SDLFunc("TTF_RenderUTF8_Solid_Wrapped", [_P(TTF_Font), c_char_p, SDL_Color, Uint32], _P(SDL_Surface), added='2.0.18'),
    SDLFunc("TTF_RenderUNICODE_Solid_Wrapped", [_P(TTF_Font), _P(Uint16), SDL_Color, Uint32], _P(SDL_Surface), added='2.0.18'),
    SDLFunc("TTF_RenderGlyph_Solid", [_P(TTF_Font), Uint16, SDL_Color], _P(SDL_Surface)),
    SDLFunc("TTF_RenderGlyph32_Solid", [_P(TTF_Font), Uint32, SDL_Color], _P(SDL_Surface), added='2.0.18'),
    SDLFunc("TTF_RenderText_Shaded", [_P(TTF_Font), c_char_p, SDL_Color, SDL_Color], _P(SDL_Surface)),
    SDLFunc("TTF_RenderUTF8_Shaded", [_P(TTF_Font), c_char_p, SDL_Color, SDL_Color], _P(SDL_Surface)),
    SDLFunc("TTF_RenderUNICODE_Shaded", [_P(TTF_Font), _P(Uint16), SDL_Color, SDL_Color], _P(SDL_Surface)),
    SDLFunc("TTF_RenderText_Shaded_Wrapped", [_P(TTF_Font), c_char_p, SDL_Color, SDL_Color, Uint32], _P(SDL_Surface), added='2.0.18'),
    SDLFunc("TTF_RenderUTF8_Shaded_Wrapped", [_P(TTF_Font), c_char_p, SDL_Color, SDL_Color, Uint32], _P(SDL_Surface), added='2.0.18'),
    SDLFunc("TTF_RenderUNICODE_Shaded_Wrapped", [_P(TTF_Font), _P(Uint16), SDL_Color, SDL_Color, Uint32], _P(SDL_Surface), added='2.0.18'),
    SDLFunc("TTF_RenderGlyph_Shaded", [_P(TTF_Font), Uint16, SDL_Color, SDL_Color], _P(SDL_Surface)),
    SDLFunc("TTF_RenderGlyph32_Shaded", [_P(TTF_Font), Uint32, SDL_Color, SDL_Color], _P(SDL_Surface), added='2.0.18'),
    SDLFunc("TTF_RenderText_Blended", [_P(TTF_Font), c_char_p, SDL_Color], _P(SDL_Surface)),
    SDLFunc("TTF_RenderUTF8_Blended", [_P(TTF_Font), c_char_p, SDL_Color], _P(SDL_Surface)),
    SDLFunc("TTF_RenderUNICODE_Blended", [_P(TTF_Font), _P(Uint16), SDL_Color], _P(SDL_Surface)),
    SDLFunc("TTF_RenderText_Blended_Wrapped", [_P(TTF_Font), c_char_p, SDL_Color, Uint32], _P(SDL_Surface)),
    SDLFunc("TTF_RenderUTF8_Blended_Wrapped", [_P(TTF_Font), c_char_p, SDL_Color, Uint32], _P(SDL_Surface)),
    SDLFunc("TTF_RenderUNICODE_Blended_Wrapped", [_P(TTF_Font), _P(Uint16), SDL_Color, Uint32], _P(SDL_Surface)),
    SDLFunc("TTF_RenderGlyph_Blended", [_P(TTF_Font), Uint16, SDL_Color], _P(SDL_Surface)),
    SDLFunc("TTF_RenderGlyph32_Blended", [_P(TTF_Font), Uint32, SDL_Color], _P(SDL_Surface), added='2.0.18'),
    SDLFunc("TTF_SetDirection", [c_int], c_int, added='2.0.18'),
    SDLFunc("TTF_SetScript", [c_int], c_int, added='2.0.18'),
    SDLFunc("TTF_CloseFont", [_P(TTF_Font)]),
    SDLFunc("TTF_Quit"),
    SDLFunc("TTF_WasInit", None, c_int),
    SDLFunc("TTF_GetFontKerningSize", [_P(TTF_Font), c_int, c_int], c_int),
    SDLFunc("TTF_GetFontKerningSizeGlyphs", [_P(TTF_Font), Uint16, Uint16], c_int, added='2.0.14'),
    SDLFunc("TTF_GetFontKerningSizeGlyphs32", [_P(TTF_Font), Uint32, Uint32], c_int, added='2.0.18'),
    SDLFunc("TTF_SetFontSDF", [_P(TTF_Font), SDL_bool], c_int, added='2.0.18'),
    SDLFunc("TTF_GetFontSDF", [_P(TTF_Font)], SDL_bool, added='2.0.18'),
]
_funcs = {}
for f in _funcdefs:
    _funcs[f.name] = _bind(f.name, f.args, f.returns, f.added)
    __all__.append(f.name)


# Python wrapper functions

def TTF_Linked_Version():
    """Gets the version of the dynamically-linked **SDL2_ttf** library.

    Returns:
        POINTER(:obj:`SDL_version`): A pointer to a structure containing the
        version of the SDL2_ttf library currently in use.

    """
    return _funcs["TTF_Linked_Version"]()

def TTF_GetFreeTypeVersion(major, minor, patch):
    """Gets the version of the FreeType library currently linked by SDL2_ttf.

    This function returns the version numbers by reference, meaning that
    it needs to be called using pre-allocated ctypes variables (see
    :func:`TTF_GlyphMetrics` for an example).

    `Note: Added in SDL_ttf 2.0.18`

    Args:
        major (byref(:obj:`~ctypes.c_int`)): A pointer to an integer containing
            the major version number of the linked FreeType library.
        minor (byref(:obj:`~ctypes.c_int`)): A pointer to an integer containing
            the minor version number of the linked FreeType library.
        patch (byref(:obj:`~ctypes.c_int`)): A pointer to an integer containing
            the patch level of the linked FreeType library.

    """
    return _funcs["TTF_GetFreeTypeVersion"](major, minor, patch)

def TTF_GetHarfBuzzVersion(major, minor, patch):
    """Gets the version of the HarfBuzz library currently linked by SDL2_ttf.

    This function returns the version numbers by reference, meaning that
    it needs to be called using pre-allocated ctypes variables (see
    :func:`TTF_GlyphMetrics` for an example).

    `Note: Added in SDL_ttf 2.0.18`

    Args:
        major (byref(:obj:`~ctypes.c_int`)): A pointer to an integer containing
            the major version number of the linked HarfBuzz library.
        minor (byref(:obj:`~ctypes.c_int`)): A pointer to an integer containing
            the minor version number of the linked HarfBuzz library.
        patch (byref(:obj:`~ctypes.c_int`)): A pointer to an integer containing
            the patch level of the linked HarfBuzz library.

    """
    return _funcs["TTF_GetHarfBuzzVersion"](major, minor, patch)

def TTF_ByteSwappedUNICODE(swapped):
    """Tells the library whether UCS-2 unicode text is generally byteswapped.
   
    A unicode BOM character in a string will override this setting for the
    remainder of that string. The default mode is non-swapped, native
    endianness of the CPU. 

    Note that this only affects the behaviour of ``UNICODE`` (UCS-2) functions
    and has no effect on UTF8 functions.
   
    Args:
        swapped (int): If 0, native CPU endianness will be used. If not 0,
            UCS-2 data will be byte-swapped relative to native CPU endianness.

    """
    return _funcs["TTF_ByteSwappedUNICODE"](swapped)

def TTF_Init():
    """Initializes the TTF engine.
   
    This function must be called before using other functions in this library
    (except for :func:`TTF_WasInit`). SDL does not have to be initialized
    before this function is called. 
   
    Returns:
        int: 0 if successful, or -1 on error.

    """
    return _funcs["TTF_Init"]()


def TTF_OpenFont(file, ptsize):
    """Opens a font file at a given size.
    
    Point sizes are based on a DPI of 72. Use the :func:`TTF_GetError` function
    to check for any errors opening the font.
   
    Args:
        file (bytes): A UTF8-encoded bytestring containing the path of the font
            file to load.
        ptsize (int): The size (in points) at which to open the font.
   
    Returns:
        POINTER(:obj:`TTF_Font`): A pointer to the opened font object, or a null
        pointer if there was an error.

    """
    return _funcs["TTF_OpenFont"](file, ptsize)

def TTF_OpenFontIndex(file, ptsize, index):
    """Opens a specific font face by index from a file at a given size.
    
    This function allows for loading a specific font face from a multi-face
    font. See :func:`TTF_OpenFont` for more information.
   
    Args:
        file (bytes): A UTF8-encoded bytestring containing the path of the font
            file to load.
        ptsize (int): The size (in points) at which to open the font.
        index (int): The index (from 0 to 15) of the font face to open. For
            font files with only one face, this should always be 0.
   
    Returns:
        POINTER(:obj:`TTF_Font`): A pointer to the opened font object, or a null
        pointer if there was an error.

    """
    return _funcs["TTF_OpenFontIndex"](file, ptsize, index)

def TTF_OpenFontRW(src, freesrc, ptsize):
    """Opens a font from a file object at a given size.

    Point sizes are based on a DPI of 72. Use the :func:`TTF_GetError` function
    to check for any errors opening the font.

    .. note::
       The file object used to create the font (``src``) must be kept in
       memory until you are done with the font. Once the ``src`` has been freed,
       performing any additional operations with the returned :obj:`TTF_Font`
       will result in a hard Python crash (segmentation fault).
   
    Args:
        src (:obj:`SDL_RWops`): A file object containing a valid font.
        freesrc (int): If non-zero, the provided file object will be closed and
            freed automatically when the resulting :obj:`TTF_Font` is closed (or
            if an error is encountered opening the font).
        ptsize (int): The size (in points) at which to open the font.
   
    Returns:
        POINTER(:obj:`TTF_Font`): A pointer to the opened font object, or a null
        pointer if there was an error.

    """
    return _funcs["TTF_OpenFontRW"](src, freesrc, ptsize)

def TTF_OpenFontIndexRW(src, freesrc, ptsize, index):
    """Opens a specific font face by index from a file object at a given size.
    
    This function allows for loading a specific font face from a multi-face
    font. See :func:`TTF_OpenFontRW` for more information.
   
    Args:
        src (:obj:`SDL_RWops`): A file object containing a valid font.
        freesrc (int): If non-zero, the provided file object will be closed and
            freed automatically when the resulting :obj:`TTF_Font` is closed (or
            if an error is encountered opening the font).
        ptsize (int): The size (in points) at which to open the font.
        index (int): The index (from 0 to 15) of the font face to open. For
            font files with only one face, this should always be 0.
   
    Returns:
        POINTER(:obj:`TTF_Font`): A pointer to the opened font object, or a null
        pointer if there was an error.

    """
    return _funcs["TTF_OpenFontIndexRW"](src, freesrc, ptsize, index)

def TTF_OpenFontDPI(file, ptsize, hdpi, vdpi):
    """Opens a font file at a given size and DPI.
    
    The font will be opened with the given horizontal and vertical target
    resolutions (in DPI). DPI scaling only applies to scalable fonts (e.g.
    TrueType). Use the :func:`TTF_GetError` function to check for any errors
    opening the font.

    `Note: Added in SDL_ttf 2.0.18`
   
    Args:
        file (bytes): A UTF8-encoded bytestring containing the path of the font
            file to load.
        ptsize (int): The size (in points) at which to open the font.
        hdpi (int): The horizontal resolution (in DPI) at which to open the font.
        vdpi (int): The vertical resolution (in DPI) at which to open the font.
   
    Returns:
        POINTER(:obj:`TTF_Font`): A pointer to the opened font object, or a null
        pointer if there was an error.

    """
    return _funcs["TTF_OpenFontDPI"](file, ptsize, hdpi, vdpi)

def TTF_OpenFontIndexDPI(file, ptsize, index, hdpi, vdpi):
    """Opens a specific font face by index from a file at a given size and DPI.
    
    See :func:`TTF_OpenFontDPI` and `:func:`TTF_OpenFontIndex` for more
    information.

    `Note: Added in SDL_ttf 2.0.18`
   
    Args:
        file (bytes): A UTF8-encoded bytestring containing the path of the font
            file to load.
        ptsize (int): The size (in points) at which to open the font.
        index (int): The index (from 0 to 15) of the font face to open. For
            font files with only one face, this should always be 0.
        hdpi (int): The horizontal resolution (in DPI) at which to open the font.
        vdpi (int): The vertical resolution (in DPI) at which to open the font.
   
    Returns:
        POINTER(:obj:`TTF_Font`): A pointer to the opened font object, or a null
        pointer if there was an error.

    """
    return _funcs["TTF_OpenFontIndexDPI"](file, ptsize, index, hdpi, vdpi)

def TTF_OpenFontDPIRW(src, freesrc, ptsize, hdpi, vdpi):
    """Opens a font from a file object at a given size and DPI.

    See :func:`TTF_OpenFontDPI` and `:func:`TTF_OpenFontRW` for more
    information.

    `Note: Added in SDL_ttf 2.0.18`
   
    Args:
        src (:obj:`SDL_RWops`): A file object containing a valid font.
        freesrc (int): If non-zero, the provided file object will be closed and
            freed automatically when the resulting :obj:`TTF_Font` is closed (or
            if an error is encountered opening the font).
        ptsize (int): The size (in points) at which to open the font.
        hdpi (int): The horizontal resolution (in DPI) at which to open the font.
        vdpi (int): The vertical resolution (in DPI) at which to open the font.
   
    Returns:
        POINTER(:obj:`TTF_Font`): A pointer to the opened font object, or a null
        pointer if there was an error.

    """
    return _funcs["TTF_OpenFontDPIRW"](src, freesrc, ptsize, hdpi, vdpi)

def TTF_OpenFontIndexDPIRW(src, freesrc, ptsize, index, hdpi, vdpi):
    """Opens a font face by index from a file object at a given size and DPI.
    
    See :func:`TTF_OpenFontDPI` and `:func:`TTF_OpenFontIndexRW` for more
    information.

    `Note: Added in SDL_ttf 2.0.18`
   
    Args:
        src (:obj:`SDL_RWops`): A file object containing a valid font.
        freesrc (int): If non-zero, the provided file object will be closed and
            freed automatically when the resulting :obj:`TTF_Font` is closed (or
            if an error is encountered opening the font).
        ptsize (int): The size (in points) at which to open the font.
        index (int): The index (from 0 to 15) of the font face to open. For
            font files with only one face, this should always be 0.
        hdpi (int): The horizontal resolution (in DPI) at which to open the font.
        vdpi (int): The vertical resolution (in DPI) at which to open the font.
   
    Returns:
        POINTER(:obj:`TTF_Font`): A pointer to the opened font object, or a null
        pointer if there was an error.

    """
    return _funcs["TTF_OpenFontIndexDPIRW"](src, freesrc, ptsize, index, hdpi, vdpi)

def TTF_SetFontSize(font, ptsize):
    """Changes the size of a TTF font object dynamically.

    Use :func:`TTF_GetError` to check for errors.

    `Note: Added in SDL_ttf 2.0.18`

    Args:
        font (:obj:`TTF_Font`): The font object to update.
        ptsize (int): The new size (in points) to use for the font.

    Returns:
        int: 0 on success, or -1 on error.

    """
    return _funcs["TTF_SetFontSize"](font, ptsize)

def TTF_SetFontSizeDPI(font, ptsize, hdpi, vdpi):
    """Changes the size and DPI of a TTF font object dynamically.

    Use :func:`TTF_GetError` to check for errors.

    `Note: Added in SDL_ttf 2.0.18`

    Args:
        font (:obj:`TTF_Font`): The font object to update.
        ptsize (int): The new size (in points) to use for the font.
        hdpi (int): The new horizontal resolution (in DPI) at which to render
            the font.
        vdpi (int): The vertical resolution (in DPI) at which to render the
            font.

    Returns:
        int: 0 on success, or -1 on error.

    """
    return _funcs["TTF_SetFontSizeDPI"](font, ptsize, hdpi, vdpi)


def TTF_GetFontStyle(font):
    """Retrieves the current rendering style of a given font.

    To check for the presence of a given style within a font, the return value
    of this function can be used with a bitwise ``&`` operator::

      style = TTF_GetFontStyle(font)
      is_bold = style & TTF_STYLE_BOLD == TTF_STYLE_BOLD
      is_underlined = style & TTF_STYLE_UNDERLINE == TTF_STYLE_UNDERLINE

    Args:
        font (:obj:`TTF_Font`): The font object for which the style should be
            retrieved.
    
    Returns:
        int: A bitmask of one or more style constants (see
        :func:`TTF_SetFontStyle`).

    """
    return _funcs["TTF_GetFontStyle"](font)

def TTF_SetFontStyle(font, style):
    """Sets the rendering style for a given font.
    
    Font styles can be specified using the following constants:

    ============= ===========================
    Style         Constant
    ============= ===========================
    Normal        ``TTF_STYLE_NORMAL``
    Bold          ``TTF_STYLE_BOLD``
    Italics       ``TTF_STYLE_ITALICS``
    Underlined    ``TTF_STYLE_UNDERLINE``
    Strikethrough ``TTF_STYLE_STRIKETHROUGH``
    ============= ===========================

    Multiple font styles (e.g. bold and italics) can be combined using the
    bitwise ``|`` operator::
 
      underlined_bold = (TTF_STYLE_BOLD | TTF_STYLE_UNDERLINE)
      TTF_SetFontStyle(font, underlined_bold)
 
    .. note::
       Setting the underline style for a font may cause the surfaces created by
       :obj:`TTF_RenderGlyph` functions to be taller, in order to make room for
       the underline to be drawn underneath.
   
    Args:
        font (:obj:`TTF_Font`): The font object for which the style will be set.
        style (int): A bitmask specifying the style(s) to use for the font.

    """
    return _funcs["TTF_SetFontStyle"](font, style)

def TTF_GetFontOutline(font):
    """Retrieves the current outline thickness of a given font.

    Args:
        font (:obj:`TTF_Font`): The font object for which the outline thickness
            should be retrieved.
    
    Returns:
        int: The outline thickness (in pixels) of the font.

    """
    return _funcs["TTF_GetFontOutline"](font)

def TTF_SetFontOutline(font, outline):
    """Sets the outline thickness of a given font.
    
    If the outline is set to zero, outlining will be disabled for the font.
   
    Args:
        font (:obj:`TTF_Font`): The font object for which the outline thickness
            will be set.
        outline (int): The new outline thickness (in pixels) for the font.

    """
    return _funcs["TTF_SetFontOutline"](font, outline)

def TTF_GetFontHinting(font):
    """Retrieves the current hinting style of a given font.

    This function returns one of the constants specified in the documentation
    for :func:`TTF_SetFontHinting`.

    Args:
        font (:obj:`TTF_Font`): The font object for which the hinting style
            should be retrieved.
    
    Returns:
        int: A constant indicating the hinting style of the font.

    """
    return _funcs["TTF_GetFontHinting"](font)

def TTF_SetFontHinting(font, hinting):
    """Sets the rendering hinting mode for a given font.
    
    The hinting mode can be specified using one of the following constants:

    ================ ==============================
    Hinting type     Constant
    ================ ==============================
    Normal           ``TTF_HINTING_NORMAL``
    Light            ``TTF_HINTING_LIGHT``
    Mono             ``TTF_HINTING_MONO``
    None             ``TTF_HINTING_NONE``
    Light (Subpixel) ``TTF_HINTING_LIGHT_SUBPIXEL``
    ================ ==============================

    Note that the ``TTF_HINTING_LIGHT_SUBPIXEL`` hinting mode requires SDL_ttf
    2.0.18 or newer. If no hinting mode is is explicitly set, "normal" hinting
    is used for rendering.
   
    Args:
        font (:obj:`TTF_Font`): The font object for which the hinting style
            will be set.
        hinting (int): A constant specifiying the hinting style to use when
            rendering text.

    """
    return _funcs["TTF_SetFontHinting"](font, hinting)


def TTF_FontHeight(font):
    """Gets the maximum pixel height of all glyphs in a given font.

    You can use this height for rendering text as close together vertically
    as possible, though adding at least one pixel height to it will space it
    so they can't touch.
   
    Args:
        font (:obj:`TTF_Font`): The font object for which the maximum glyph
            height should be retrieved.

    Returns:
        int: The maximum height (in pixels) of all glyphs in the font.

    """
    return _funcs["TTF_FontHeight"](font)

def TTF_FontAscent(font):
    """Gets the maximum pixel ascent of all glyphs in a given font.

    The ascent of a glyph is defined as the distance from the top of the
    glyph to its baseline. For example, a lower-case "t" will generally have a
    larger ascent than a lower-case "o".
   
    Args:
        font (:obj:`TTF_Font`): The font object for which the maximum glyph
            ascent should be retrieved.

    Returns:
        int: The maximum ascent (in pixels) of all glyphs in the font.

    """
    return _funcs["TTF_FontAscent"](font)

def TTF_FontDescent(font):
    """Gets the maximum pixel descent of all glyphs in a given font.

    The descent of a glyph is defined as the distance from the bottom of the
    glyph to its baseline. For example, a lower-case "g" will typically have a
    non-zero descent whereas a lower-case "o" will generally have a descent of
    zero.
   
    Args:
        font (:obj:`TTF_Font`): The font object for which the maximum glyph
            descent should be retrieved.

    Returns:
        int: The maximum descent (in pixels) of all glyphs in the font.

    """
    return _funcs["TTF_FontDescent"](font)

def TTF_FontLineSkip(font):
    """Gets the recommended spacing between lines for a given font.

    This is usually larger than the result of :func:`TTF_FontHeight`.
   
    Args:
        font (:obj:`TTF_Font`): The font object for which the recommended line
            skip height should be retrieved.

    Returns:
        int: The recommended line skip height (in pixels) for the font.

    """
    return _funcs["TTF_FontLineSkip"](font)

def TTF_GetFontKerning(font):
    """Checks whether or not kerning is enabled for a given font.
   
    Args:
        font (:obj:`TTF_Font`): The font object for which the kerning status
            should be retrieved.

    Returns:
        int: Non-zero if kerning is enabled for the font, otherwise 0.

    """
    return _funcs["TTF_GetFontKerning"](font)

def TTF_SetFontKerning(font, allowed):
    """Enables or disables kerning for a given font.

    Kerning is enabled for all fonts by default.
   
    Args:
        font (:obj:`TTF_Font`): The font object for which the kerning status
            should be changed.
        allowed (int): 0 to disable kerning, or non-zero to enable it.

    """
    return _funcs["TTF_SetFontKerning"](font, allowed)

def TTF_FontFaces(font):
    """Gets the number of faces ("sub-fonts") available in a given font.
   
    This is a count of the number of specific fonts (based on size and style
    and other typographical features) contained in the font itself.

    Args:
        font (:obj:`TTF_Font`): The font object for which the number of faces
            should be retrieved.

    Returns:
        int: The number of faces in the font.

    """
    return _funcs["TTF_FontFaces"](font)

def TTF_FontFaceIsFixedWidth(font):
    """Checks if the current face of a given font is fixed width.

    Fixed width fonts are monospace, meaning every character that exists in the
    font is the same width, thus you can assume that a rendered string's width
    is going to be the result of a simple calculation:
    ``glyph_width * string_length``.

    Args:
        font (:obj:`TTF_Font`): The font object for which the fixed-width status
            should be retrieved.

    Returns:
        int: A positive integer if the font is fixed-width, otherwise 0.
   
   """
    return _funcs["TTF_FontFaceIsFixedWidth"](font)

def TTF_FontFaceFamilyName(font):
    """Gets the current family name (e.g. Helvetica) from a given font.

    Args:
        font (:obj:`TTF_Font`): The font object for which the family name
            should be retrieved.

    Returns:
        bytes: The family name of the given font, or ``None`` if not
        available.

    """
    return _funcs["TTF_FontFaceFamilyName"](font)

def TTF_FontFaceStyleName(font):
    """Gets the current style name (e.g. Bold) from a given font.

    Args:
        font (:obj:`TTF_Font`): The font object for which the style name
            should be retrieved.

    Returns:
        bytes: The style name of the given font, or ``None`` if not
        available.

    """
    return _funcs["TTF_FontFaceStyleName"](font)

def TTF_GlyphIsProvided(font, ch):
    """Checks whether a character is provided by a given font.

    The built-in Python :func:`ord` function can be used to convert a string
    character to an integer for use with this function (e.g. ``ord("A")``).

    Args:
        font (:obj:`TTF_Font`): The font object within which to check for a
            glyph for the given character.
        ch (int): A unicode integer representing the character to check for
            within the font.

    Returns:
        int: A non-zero integer if the font provides a glyph for the character,
        otherwise 0.

    """
    return _funcs["TTF_GlyphIsProvided"](font, ch)

def TTF_GlyphIsProvided32(font, ch):
    """Checks whether a character is provided by a given font.

    Functionally identical to :func:`TTF_GlyphIsProvided`, except it supports
    32-bit character codes instead of just 16-bit ones.

    `Note: Added in SDL_ttf 2.0.18`

    """
    return _funcs["TTF_GlyphIsProvided32"](font, ch)

def TTF_GlyphMetrics(font, ch, minx, maxx, miny, maxy, advance):
    """Gets the glyph metrics for a character with a given font.
    
    This function returns the calculated metrics by reference, meaning that
    it needs to be called using pre-allocated ctypes variables::

       from ctypes import c_int, byref

       minX, maxX, minY, maxY = c_int(0), c_int(0), c_int(0), c_int(0)
       adv = c_int(0)
       TTF_GlyphMetrics(
           font, ord(char),
           byref(minX), byref(maxX), byref(minY), byref(maxY), byref(adv)
       )
       results = [x.value for x in (minX, maxX, minY, maxY, adv)]

    The following link may be useful for understanding what these metrics mean:
    http://freetype.sourceforge.net/freetype2/docs/tutorial/step2.html

    Args:
        font (:obj:`TTF_Font`): The font object to use.
        ch (int): A unicode integer representing the character for which to
            retrieve glyph metrics.
        minx (byref(:obj:`~ctypes.c_int`)): A pointer to an integer in which to
            store the glyph's minimum X offset.
        maxx (byref(:obj:`~ctypes.c_int`)): A pointer to an integer in which to
            store the glyph's maximum X offset.
        miny (byref(:obj:`~ctypes.c_int`)): A pointer to an integer in which to
            store the glyph's minimum Y offset.
        maxy (byref(:obj:`~ctypes.c_int`)): A pointer to an integer in which to
            store the glyph's maximum Y offset.

    Returns:
        int: 0 on success, or -1 on error (e.g. if the glyph does not exist in
        the font). 

    """
    return _funcs["TTF_GlyphMetrics"](font, ch, minx, maxx, miny, maxy, advance)

def TTF_GlyphMetrics32(font, ch, minx, maxx, miny, maxy, advance):
    """Gets the glyph metrics for a character with a given font.

    Functionally identical to :func:`TTF_GlyphMetrics`, except it supports
    32-bit character codes instead of just 16-bit ones.

    `Note: Added in SDL_ttf 2.0.18`

    """
    return _funcs["TTF_GlyphMetrics32"](font, ch, minx, maxx, miny, maxy, advance)


def TTF_SizeText(font, text, w, h):
    """Calculates the size of an ASCII string rendered with a given font.

    This function does not perform any actual rendering, but correct kerning is
    performed to get the actual width. For a string without any newlines, the
    height will be the same as that returned by :func:`TTF_FontHeight`.

    This function returns the calculated metrics by reference, meaning that
    it needs to be called using pre-allocated ctypes variables::

       from ctypes import c_int, byref

       text_w, text_h = c_int(0), c_int(0)
       TTF_SizeText(font, b"hello!", byref(text_w), byref(text_h))
       text_size = [x.value for x in (text_w, text_h)]

    Args:
        font (:obj:`TTF_Font`): The font object to use.
        text (bytes): An ASCII-encoded bytestring of text for which the rendered
            surface size should be calculated.
        w (byref(:obj:`~ctypes.c_int`)): A pointer to an integer in which to
            store the calculated surface width (in pixels).
        h (byref(:obj:`~ctypes.c_int`)): A pointer to an integer in which to
            store the calculated surface height (in pixels).

    Returns:
        int: 0 on success, or -1 on error (e.g. if a glyph is not found in
        the font).

    """
    return _funcs["TTF_SizeText"](font, text, w, h)

def TTF_SizeUTF8(font, text, w, h):
    """Calculates the size of a UTF8-encoded string rendered with a given font.

    See :func:`TTF_SizeText` for more info.

    Args:
        font (:obj:`TTF_Font`): The font object to use.
        text (bytes): A UTF8-encoded bytestring of text for which the rendered
            surface size should be calculated.
        w (byref(:obj:`~ctypes.c_int`)): A pointer to an integer in which to
            store the calculated surface width (in pixels).
        h (byref(:obj:`~ctypes.c_int`)): A pointer to an integer in which to
            store the calculated surface height (in pixels).

    Returns:
        int: 0 on success, or -1 on error (e.g. if a glyph is not found in
        the font).

    """
    return _funcs["TTF_SizeUTF8"](font, text, w, h)

def TTF_SizeUNICODE(font, text, w, h):
    """Calculates the size of a UCS-2 encoded string rendered with a given font.

    See :func:`TTF_SizeText` and :func:`TTF_RenderUNICODE_Solid` for more info.

    Args:
        font (:obj:`TTF_Font`): The font object to use.
        text (byref(:obj:`~ctypes.c_uint16`)): A ctypes array containing a UCS-2
            encoded string of text for which the rendered surface size should be
            calculated.
        w (byref(:obj:`~ctypes.c_int`)): A pointer to an integer in which to
            store the calculated surface width (in pixels).
        h (byref(:obj:`~ctypes.c_int`)): A pointer to an integer in which to
            store the calculated surface height (in pixels).

    Returns:
        int: 0 on success, or -1 on error (e.g. if a glyph is not found in
        the font).

    """
    return _funcs["TTF_SizeUNICODE"](font, text, w, h)

def TTF_MeasureText(font, text, measure_width, extent, count):
    """Gets the number of characters that can fit within a given width.

    This function determines how many rendered characters from a given string of
    ASCII text can fit within a given width. It additionally returns the
    rendered width of the fitting characters.

    Use the :func:`TTF_GetError` function to check for any errors.

    `Note: Added in SDL_ttf 2.0.18`

    Args:
        font (:obj:`TTF_Font`): The font object to use.
        text (bytes): An ASCII-encoded bytestring of text.
        measure_width (int): The maximum width (in pixels) of the rendered
            output surface.
        extent (byref(:obj:`~ctypes.c_int`)): A pointer to an integer containing
            the calculated rendered width of the first ``count`` characters of
            the string.
        count (byref(:obj:`~ctypes.c_int`)): A pointer to an integer containing
            the returned number of characters from the string that can fit
            within the given width.

    Returns:
        int: 0 on success, or -1 on error (e.g. if a glyph is not found in
        the font).

    """
    return _funcs["TTF_MeasureText"](font, text, measure_width, extent, count)

def TTF_MeasureUTF8(font, text, measure_width, extent, count):
    """Gets the number of characters that can fit within a given width.

    Identical to :func:`TTF_MeasureText`, except that this function is used with
    UTF8-encoded bytestrings instead of ASCII-encoded ones.

    `Note: Added in SDL_ttf 2.0.18`

    Args:
        font (:obj:`TTF_Font`): The font object to use.
        text (bytes): A UTF8-encoded bytestring of text.
        measure_width (int): The maximum width (in pixels) of the rendered
            output surface.
        extent (byref(:obj:`~ctypes.c_int`)): A pointer to an integer containing
            the calculated rendered width of the first ``count`` characters of
            the string.
        count (byref(:obj:`~ctypes.c_int`)): A pointer to an integer containing
            the returned number of characters from the string that can fit
            within the given width.

    Returns:
        int: 0 on success, or -1 on error (e.g. if a glyph is not found in
        the font).

    """
    return _funcs["TTF_MeasureUTF8"](font, text, measure_width, extent, count)

def TTF_MeasureUNICODE(font, text, measure_width, extent, count):
    """Gets the number of characters that can fit within a given width.

    Identical to :func:`TTF_MeasureText`, except that this function is used with
    UCS-2 byte arrays instead of ASCII-encoded bytestrings. See the
    :func:`TTF_RenderUNICODE_Solid` documentation for more information on
    converting text to UCS-2 in Python.

    `Note: Added in SDL_ttf 2.0.18`

    Args:
        font (:obj:`TTF_Font`): The font object to use.
        text (byref(:obj:`~ctypes.c_uint16`)): A ctypes array containing a UCS-2
            encoded string of text.
        measure_width (int): The maximum width (in pixels) of the rendered
            output surface.
        extent (byref(:obj:`~ctypes.c_int`)): A pointer to an integer containing
            the calculated rendered width of the first ``count`` characters of
            the string.
        count (byref(:obj:`~ctypes.c_int`)): A pointer to an integer containing
            the returned number of characters from the string that can fit
            within the given width.

    Returns:
        int: 0 on success, or -1 on error (e.g. if a glyph is not found in
        the font).

    """
    return _funcs["TTF_MeasureUNICODE"](font, text, measure_width, extent, count)


def TTF_RenderText_Solid(font, text, fg):
    """Renders an ASCII-encoded string to a non-antialiased 8-bit surface.

    The ``Solid`` family of TTF functions render text to an 8-bit palettized
    :obj:`SDL_Surface` with a transparent background and no antialiasing.
    This is the fastest (and lowest quality) of all TTF rendering types.

    The resulting surface has a transparent background unlike
    :func:`TTF_RenderText_Shaded`, but the rendered text is not antialised and
    will thus appear pixelated and difficult to read at small sizes. The
    resulting surface should blit faster than the one returned by
    :func:`TTF_RenderText_Blended`. This rendering type should be used in cases
    when you need to render lots of text very quickly (e.g. if you're updating
    it every frame) or when you don't care about antialiasing.
   
    The 0 pixel is the colorkey for the resulting surface, giving a transparent
    background, and the 1 pixel is set to the text color. This allows you to
    change the color without having to render the text again. Palette index 0
    is not drawn when the returned surface is blitted to another surface (since
    it is the colorkey and thus transparent), though its actual color is 255
    minus each of the RGB components of the foreground color.
   
    Args:
        font (:obj:`TTF_Font`): The font object to use.
        text (bytes): An ASCII-encoded bytestring of text to render.
        fg (:obj:`SDL_Color`): The color to use for rendering the text. This
            becomes colormap index 1.

    Returns:
        POINTER(:obj:`SDL_Surface`): A pointer to the new surface containing the
        rendered text, or a null pointer if there was an error.

    """
    return _funcs["TTF_RenderText_Solid"](font, text, fg)

def TTF_RenderUTF8_Solid(font, text, fg):
    """Renders a UTF8-encoded string to a non-antialiased 8-bit surface.

    See :func:`TTF_RenderText_Solid` for more details on the rendering style.

    Args:
        font (:obj:`TTF_Font`): The font object to use.
        text (bytes): A UTF8-encoded bytestring of text to render.
        fg (:obj:`SDL_Color`): The color to use for rendering the text. This
            becomes colormap index 1.

    Returns:
        POINTER(:obj:`SDL_Surface`): A pointer to the new surface containing the
        rendered text, or a null pointer if there was an error.

    """
    return _funcs["TTF_RenderUTF8_Solid"](font, text, fg)

def TTF_RenderUNICODE_Solid(font, text, fg):
    """Renders a UCS-2 encoded string to a non-antialiased 8-bit surface.

    The required text input format for this function is a ctypes array of
    UNICODE (UCS-2) glyphs in uint16 format, optionally terminated by a
    byte-order mark (``UNICODE_BOM_NATIVE`` or ``UNICODE_BOM_SWAPPED``)
    indicating how the text should be interpreted. Python strings can be
    converted to this format using the following process::

       # Generate a UCS-2 array from a Python string
       teststr = u"Hello world!"
       strlen = len(teststr + 1)  # +1 for byte-order mark
       intstr = unpack('H' * strlen, teststr.encode('utf-16'))
       intstr = intstr + (0, )  # Null-terminate the string
       strarr = (ctypes.c_uint16 * len(intstr))(*intstr)

       # Render the UCS-2 string
       col = SDL_Color(0, 0, 0)
       rendered = TTF_RenderUNICODE_Solid(font, strarr, col)

    Unless there is a very specific need, the ``TTF_RenderUTF8`` functions should
    always be used instead of their ``TTF_RenderUNICODE`` counterparts. In
    addition to having a much friendlier Python API, SDL_ttf uses the
    ``TTF_RenderUTF8`` functions internally for all the ``TTF_RenderUNICODE``
    functions anyway so there is no benefit in terms of supporting a wider range
    of characters.

    See :func:`TTF_RenderText_Solid` for more details on the rendering style.

    Args:
        font (:obj:`TTF_Font`): The font object to use.
        text (byref(:obj:`~ctypes.c_uint16`)): A ctypes array containing a UCS-2
            encoded string of text to render.
        fg (:obj:`SDL_Color`): The color to use for rendering the text. This
            becomes colormap index 1.

    Returns:
        POINTER(:obj:`SDL_Surface`): A pointer to the new surface containing the
        rendered text, or a null pointer if there was an error.

    """
    return _funcs["TTF_RenderUNICODE_Solid"](font, text, fg)

def TTF_RenderText_Solid_Wrapped(font, text, fg, wrapLength):
    """Renders an ASCII-encoded string to a non-antialiased 8-bit surface.

    This function is identical to :func:`TTF_RenderText_Solid`, except that
    any lines exceeding the specified wrap length will be wrapped to fit within
    the given width.

    `Note: Added in SDL_ttf 2.0.18`
   
    Args:
        font (:obj:`TTF_Font`): The font object to use.
        text (bytes): An ASCII-encoded bytestring of text to render.
        fg (:obj:`SDL_Color`): The color to use for rendering the text.
        wrapLength (int): The maximum width of the output surface (in pixels)

    Returns:
        POINTER(:obj:`SDL_Surface`): A pointer to the new surface containing the
        rendered text, or a null pointer if there was an error.

    """
    return _funcs["TTF_RenderText_Solid_Wrapped"](font, text, fg, wrapLength)

def TTF_RenderUTF8_Solid_Wrapped(font, text, fg, wrapLength):
    """Renders a UTF8-encoded string to a non-antialiased 8-bit surface.

    This function is identical to :func:`TTF_RenderUTF8_Solid`, except that
    any lines exceeding the specified wrap length will be wrapped to fit within
    the given width.

    `Note: Added in SDL_ttf 2.0.18`
   
    Args:
        font (:obj:`TTF_Font`): The font object to use.
        text (bytes): A UTF8-encoded bytestring of text to render.
        fg (:obj:`SDL_Color`): The color to use for rendering the text.
        wrapLength (int): The maximum width of the output surface (in pixels)

    Returns:
        POINTER(:obj:`SDL_Surface`): A pointer to the new surface containing the
        rendered text, or a null pointer if there was an error.

    """
    return _funcs["TTF_RenderUTF8_Solid_Wrapped"](font, text, fg, wrapLength)

def TTF_RenderUNICODE_Solid_Wrapped(font, text, fg, wrapLength):
    """Renders a UCS-2 encoded string to a non-antialiased 8-bit surface.

    This function is identical to :func:`TTF_RenderUNICODE_Solid`, except that
    any lines exceeding the specified wrap length will be wrapped to fit within
    the given width.

    `Note: Added in SDL_ttf 2.0.18`

    Args:
        font (:obj:`TTF_Font`): The font object to use.
        text (byref(:obj:`~ctypes.c_uint16`)): A ctypes array containing a UCS-2
            encoded string of text to render.
        fg (:obj:`SDL_Color`): The color to use for rendering the text.
        wrapLength (int): The maximum width of the output surface (in pixels)

    Returns:
        POINTER(:obj:`SDL_Surface`): A pointer to the new surface containing the
        rendered text, or a null pointer if there was an error.

    """
    return _funcs["TTF_RenderUNICODE_Solid_Wrapped"](font, text, fg, wrapLength)

def TTF_RenderGlyph_Solid(font, ch, fg):
    """Renders a unicode character to a non-antialiased 8-bit surface.

    The built-in Python :func:`ord` function can be used to convert a string
    character to an integer for use with this function (e.g. ``ord("A")``).

    The glyph is rendered without any padding or centering in the X direction,
    and is aligned normally in the Y direction. See :func:`TTF_RenderText_Solid`
    for more details on the rendering style.

    Args:
        font (:obj:`TTF_Font`): The font object to use.
        ch (int): A unicode integer representing the glyph to render.
        fg (:obj:`SDL_Color`): The color to use for rendering the glyph. This
            becomes colormap index 1.

    Returns:
        POINTER(:obj:`SDL_Surface`): A pointer to the new surface containing the
        rendered glyph, or a null pointer if there was an error.

    """
    return _funcs["TTF_RenderGlyph_Solid"](font, ch, fg)

def TTF_RenderGlyph32_Solid(font, ch, fg):
    """Renders a unicode character to a non-antialiased 8-bit surface.

    Functionally identical to :func:`TTF_RenderGlyph_Solid`, except it supports
    32-bit character codes instead of just 16-bit ones.

    `Note: Added in SDL_ttf 2.0.18`

    """
    return _funcs["TTF_RenderGlyph32_Solid"](font, ch, fg)


def TTF_RenderText_Shaded(font, text, fg, bg):
    """Renders an ASCII-encoded string to a solid antialiased 8-bit surface.

    The ``Shaded`` family of TTF functions render text to an 8-bit palettized
    :obj:`SDL_Surface` with a solid background color and antialiasing. This is
    the second-fastest of the text rendering types, being slightly faster than
    :func:`TTF_RenderText_Blended` but slower than :func:`TTF_RenderText_Solid`.

    Text rendered using the ``Shaded`` method will be antialiased, but the
    resulting surface will have a solid background colour instead of a
    transparent one. Surfaces rendered with this function should blit as quickly
    as those created with :func:`TTF_RenderText_Blended`. This rendering type
    should be used in cases when you want nice-looking text but don't need
    background transparency.

    The 0 pixel is the background color for the resulting surface, while other
    pixels have varying levels of the foreground color. This results in a box of
    the background color around the text in the foreground color.

    Args:
        font (:obj:`TTF_Font`): The font object to use.
        text (bytes): An ASCII-encoded bytestring of text to render.
        fg (:obj:`SDL_Color`): The color to use for rendering the text. This
            becomes colormap index 1.
        bg (:obj:`SDL_Color`): The background fill color for the text. This
            becomes colormap index 0.

    Returns:
        POINTER(:obj:`SDL_Surface`): A pointer to the new surface containing the
        rendered text, or a null pointer if there was an error.

    """
    return _funcs["TTF_RenderText_Shaded"](font, text, fg, bg)

def TTF_RenderUTF8_Shaded(font, text, fg, bg):
    """Renders a UTF8-encoded string to a solid antialiased 8-bit surface.

    See :func:`TTF_RenderText_Shaded` for more details on the rendering style.

    Args:
        font (:obj:`TTF_Font`): The font object to use.
        text (bytes): A UTF8-encoded bytestring of text to render.
        fg (:obj:`SDL_Color`): The color to use for rendering the text. This
            becomes colormap index 1.
        bg (:obj:`SDL_Color`): The background fill color for the text. This
            becomes colormap index 0.

    Returns:
        POINTER(:obj:`SDL_Surface`): A pointer to the new surface containing the
        rendered text, or a null pointer if there was an error.

    """
    return _funcs["TTF_RenderUTF8_Shaded"](font, text, fg, bg)

def TTF_RenderUNICODE_Shaded(font, text, fg, bg):
    """Renders a UCS-2 encoded string to a solid antialiased 8-bit surface.

    See :func:`TTF_RenderText_Shaded` for more details on the rendering style,
    and :func:`TTF_RenderUNICODE_Solid` for documentation of the text format.

    Args:
        font (:obj:`TTF_Font`): The font object to use.
        text (byref(:obj:`~ctypes.c_uint16`)): A ctypes array containing a UCS-2
            encoded string of text to render.
        fg (:obj:`SDL_Color`): The color to use for rendering the text. This
            becomes colormap index 1.
        bg (:obj:`SDL_Color`): The background fill color for the text. This
            becomes colormap index 0.

    Returns:
        POINTER(:obj:`SDL_Surface`): A pointer to the new surface containing the
        rendered text, or a null pointer if there was an error.

    """
    return _funcs["TTF_RenderUNICODE_Shaded"](font, text, fg, bg)

def TTF_RenderText_Shaded_Wrapped(font, text, fg, bg, wrapLength):
    """Renders an ASCII-encoded string to a solid antialiased 8-bit surface.

    This function is identical to :func:`TTF_RenderText_Shaded`, except that
    any lines exceeding the specified wrap length will be wrapped to fit within
    the given width.

    `Note: Added in SDL_ttf 2.0.18`
   
    Args:
        font (:obj:`TTF_Font`): The font object to use.
        text (bytes): An ASCII-encoded bytestring of text to render.
        fg (:obj:`SDL_Color`): The color to use for rendering the text. This
            becomes colormap index 1.
        bg (:obj:`SDL_Color`): The background fill color for the text. This
            becomes colormap index 0.
        wrapLength (int): The maximum width of the output surface (in pixels)

    Returns:
        POINTER(:obj:`SDL_Surface`): A pointer to the new surface containing the
        rendered text, or a null pointer if there was an error.

    """
    return _funcs["TTF_RenderText_Shaded_Wrapped"](font, text, fg, bg, wrapLength)

def TTF_RenderUTF8_Shaded_Wrapped(font, text, fg, bg, wrapLength):
    """Renders a UTF8-encoded string to a solid antialiased 8-bit surface.

    This function is identical to :func:`TTF_RenderUTF8_Shaded`, except that
    any lines exceeding the specified wrap length will be wrapped to fit within
    the given width.

    `Note: Added in SDL_ttf 2.0.18`
   
    Args:
        font (:obj:`TTF_Font`): The font object to use.
        text (byref(:obj:`~ctypes.c_uint16`)): A ctypes array containing a UCS-2
            encoded string of text to render.
        fg (:obj:`SDL_Color`): The color to use for rendering the text. This
            becomes colormap index 1.
        bg (:obj:`SDL_Color`): The background fill color for the text. This
            becomes colormap index 0.
        wrapLength (int): The maximum width of the output surface (in pixels)

    Returns:
        POINTER(:obj:`SDL_Surface`): A pointer to the new surface containing the
        rendered text, or a null pointer if there was an error.

    """
    return _funcs["TTF_RenderUTF8_Shaded_Wrapped"](font, text, fg, bg, wrapLength)

def TTF_RenderUNICODE_Shaded_Wrapped(font, text, fg, bg, wrapLength):
    """Renders a UCS-2 encoded string to a solid antialiased 8-bit surface.

    This function is identical to :func:`TTF_RenderUNICODE_Shaded`, except that
    any lines exceeding the specified wrap length will be wrapped to fit within
    the given width.

    `Note: Added in SDL_ttf 2.0.18`
   
    Args:
        font (:obj:`TTF_Font`): The font object to use.
        text (bytes): A UTF8-encoded bytestring of text to render.
        fg (:obj:`SDL_Color`): The color to use for rendering the text. This
            becomes colormap index 1.
        bg (:obj:`SDL_Color`): The background fill color for the text. This
            becomes colormap index 0.
        wrapLength (int): The maximum width of the output surface (in pixels)

    Returns:
        POINTER(:obj:`SDL_Surface`): A pointer to the new surface containing the
        rendered text, or a null pointer if there was an error.

    """
    return _funcs["TTF_RenderUNICODE_Shaded_Wrapped"](font, text, fg, bg, wrapLength)

def TTF_RenderGlyph_Shaded(font, ch, fg, bg):
    """Renders a unicode character to an 8-bit surface using a given font.

    See :func:`TTF_RenderText_Shaded` for more details on the rendering style,
    and :func:`TTF_RenderGlyph_Solid` for additional usage information.

    Args:
        font (:obj:`TTF_Font`): The font object to use.
        ch (int): A unicode integer representing the glyph to render.
        fg (:obj:`SDL_Color`): The color to use for rendering the glyph. This
            becomes colormap index 1.
        bg (:obj:`SDL_Color`): The background fill color for the glyph. This
            becomes colormap index 0.

    Returns:
        POINTER(:obj:`SDL_Surface`): A pointer to the new surface containing the
        rendered glyph, or a null pointer if there was an error.

    """
    return _funcs["TTF_RenderGlyph_Shaded"](font, ch, fg, bg)

def TTF_RenderGlyph32_Shaded(font, ch, fg, bg):
    """Renders a unicode character to an 8-bit surface using a given font.

    Functionally identical to :func:`TTF_RenderGlyph_Shaded`, except it supports
    32-bit character codes instead of just 16-bit ones.

    `Note: Added in SDL_ttf 2.0.18`

    """
    return _funcs["TTF_RenderGlyph32_Shaded"](font, ch, fg, bg)


def TTF_RenderText_Blended(font, text, fg):
    """Renders an ASCII-encoded string to an antialiased 32-bit surface.

    The ``Blended`` family of TTF functions render text to a 32-bit ARGB
    :obj:`SDL_Surface` with antialiasing and background transparency.
    This is the highest quality (and slowest) of all TTF rendering types.

    The rendered text will be antialiased on a transparent surface using alpha
    blending. This rendering type should be used in cases when you want to
    overlay the rendered text over something else, and in in most other cases
    where high performance isn't the primary concern.

    .. note::
       To render an RGBA surface instead of an ARGB one, just swap the R and B
       values when creating the SDL_Color.
   
    Args:
        font (:obj:`TTF_Font`): The font object to use.
        text (bytes): An ASCII-encoded bytestring of text to render.
        fg (:obj:`SDL_Color`): The color to use for rendering the text.

    Returns:
        POINTER(:obj:`SDL_Surface`): A pointer to the new surface containing the
        rendered text, or a null pointer if there was an error.

    """
    return _funcs["TTF_RenderText_Blended"](font, text, fg)

def TTF_RenderUTF8_Blended(font, text, fg):
    """Renders a UTF8-encoded string to an antialiased 32-bit surface.

    See :func:`TTF_RenderText_Blended` for more details on the rendering style.
   
    Args:
        font (:obj:`TTF_Font`): The font object to use.
        text (bytes): A UTF8-encoded bytestring of text to render.
        fg (:obj:`SDL_Color`): The color to use for rendering the text.

    Returns:
        POINTER(:obj:`SDL_Surface`): A pointer to the new surface containing the
        rendered text, or a null pointer if there was an error.

    """
    return _funcs["TTF_RenderUTF8_Blended"](font, text, fg)

def TTF_RenderUNICODE_Blended(font, text, fg):
    """Renders a UCS-2 encoded string to an antialiased 32-bit surface.

    See :func:`TTF_RenderText_Blended` for more details on the rendering style,
    and :func:`TTF_RenderUNICODE_Solid` for documentation of the text format.

    Args:
        font (:obj:`TTF_Font`): The font object to use.
        text (byref(:obj:`~ctypes.c_uint16`)): A ctypes array containing a UCS-2
            encoded string of text to render.
        fg (:obj:`SDL_Color`): The color to use for rendering the text.

    Returns:
        POINTER(:obj:`SDL_Surface`): A pointer to the new surface containing the
        rendered text, or a null pointer if there was an error.

    """
    return _funcs["TTF_RenderUNICODE_Blended"](font, text, fg)

def TTF_RenderText_Blended_Wrapped(font, text, fg, wrapLength):
    """Renders an ASCII-encoded string to an antialiased 32-bit surface.

    This function is identical to :func:`TTF_RenderText_Blended`, except that
    any lines exceeding the specified wrap length will be wrapped to fit within
    the given width.
   
    Args:
        font (:obj:`TTF_Font`): The font object to use.
        text (bytes): An ASCII-encoded bytestring of text to render.
        fg (:obj:`SDL_Color`): The color to use for rendering the text.
        wrapLength (int): The maximum width of the output surface (in pixels)

    Returns:
        POINTER(:obj:`SDL_Surface`): A pointer to the new surface containing the
        rendered text, or a null pointer if there was an error.

    """
    return _funcs["TTF_RenderText_Blended_Wrapped"](font, text, fg, wrapLength)

def TTF_RenderUTF8_Blended_Wrapped(font, text, fg, wrapLength):
    """Renders a UTF8-encoded string to an antialiased 32-bit surface.

    This function is identical to :func:`TTF_RenderUTF8_Blended`, except that
    any lines exceeding the specified wrap length will be wrapped to fit within
    the given width.
   
    Args:
        font (:obj:`TTF_Font`): The font object to use.
        text (bytes): A UTF8-encoded bytestring of text to render.
        fg (:obj:`SDL_Color`): The color to use for rendering the text.
        wrapLength (int): The maximum width of the output surface (in pixels)

    Returns:
        POINTER(:obj:`SDL_Surface`): A pointer to the new surface containing the
        rendered text, or a null pointer if there was an error.

    """
    return _funcs["TTF_RenderUTF8_Blended_Wrapped"](font, text, fg, wrapLength)

def TTF_RenderUNICODE_Blended_Wrapped(font, text, fg, wrapLength):
    """Renders a UCS-2 encoded string to an antialiased 32-bit surface.

    This function is identical to :func:`TTF_RenderUNICODE_Blended`, except that
    any lines exceeding the specified wrap length will be wrapped to fit within
    the given width.
   
    Args:
        font (:obj:`TTF_Font`): The font object to use.
        text (byref(:obj:`~ctypes.c_uint16`)): A ctypes array containing a UCS-2
            encoded string of text to render.
        fg (:obj:`SDL_Color`): The color to use for rendering the text.
        wrapLength (int): The maximum width of the output surface (in pixels)

    Returns:
        POINTER(:obj:`SDL_Surface`): A pointer to the new surface containing the
        rendered text, or a null pointer if there was an error.

    """
    return _funcs["TTF_RenderUNICODE_Blended_Wrapped"](font, text, fg, wrapLength)

def TTF_RenderGlyph_Blended(font, ch, fg):
    """Renders a unicode character to an antialiased 32-bit surface.

    See :func:`TTF_RenderText_Blended` for more details on the rendering style,
    and :func:`TTF_RenderGlyph_Solid` for additional usage information.

    Args:
        font (:obj:`TTF_Font`): The font object to use.
        ch (int): A unicode integer representing the glyph to render.
        fg (:obj:`SDL_Color`): The color to use for rendering the glyph.

    Returns:
        POINTER(:obj:`SDL_Surface`): A pointer to the new surface containing the
        rendered glyph, or a null pointer if there was an error.

    """
    return _funcs["TTF_RenderGlyph_Blended"](font, ch, fg)

def TTF_RenderGlyph32_Blended(font, ch, fg):
    """Renders a unicode character to an antialiased 32-bit surface.

    Functionally identical to :func:`TTF_RenderGlyph_Blended`, except it
    supports 32-bit character codes instead of just 16-bit ones.

    `Note: Added in SDL_ttf 2.0.18`

    """
    return _funcs["TTF_RenderGlyph32_Blended"](font, ch, fg)

TTF_RenderText = TTF_RenderText_Shaded
TTF_RenderUTF8 = TTF_RenderUTF8_Shaded
TTF_RenderUNICODE = TTF_RenderUNICODE_Shaded


def TTF_SetDirection(direction):
    """Sets the global text direction to use for rendering.

    This function lets you manually specify the direction in which SDL_ttf
    should render text, and can be set or changed at any time.

    The direction value is passed to the underlying HarfBuzz library, meaning
    that the direction must be one of the following HarfBuzz constants:

    =============== ====================
    Text Direction  Constant
    =============== ====================
    Left-to-right   ``HB_DIRECTION_LTR``
    Right-to-left   ``HB_DIRECTION_RTL``
    Top-to-bottom   ``HB_DIRECTION_TTB``
    Bottom-to-top   ``HB_DIRECTION_BTT``
    =============== ====================

    For convenience, these constants are provided by the ``sdl2.sdlttf`` module.
    If not specified, SDL_ttf defaults to left-to-right text rendering.

    `Note: Added in SDL_ttf 2.0.18`

    Args:
        direction (int): A constant specifying the direction to use for text
            rendering with the TTF library.

    Returns:
        int: 0 on success, or -1 if HarfBuzz not available.

    """
    return _funcs["TTF_SetDirection"](direction)

def TTF_SetScript(script):
    """Sets the global script (e.g. Arabic) to use for rendering text.

    Setting the script gives the text renderer extra information about how
    to best shape words and characters for a given language. This can produce
    better results when rendering with non-Latin languages and fonts.

    The script value is passed to the underlying HarfBuzz library, meaning that
    it needs to be specified as a HarfBuzz script constant. To make this
    convenient, the ``sdl2.sdlttf`` module implements HarfBuzz's :func:`HB_TAG`
    macro for converting ISO 15924 character codes to HarfBuzz scripts::

        arabic_script = HB_TAG('A', 'r', 'a', 'b')
        TTF_SetScript(arabic_script)

    If no script has been set, the TTF library defaults to the unknown ('Zzzz')
    script.

    `Note: Added in SDL_ttf 2.0.18`

    Args:
        script (int): An integer specifying the script style to use for text
            shaping.

    Returns:
        int: 0 on success, or -1 if HarfBuzz not available.

    """
    return _funcs["TTF_SetScript"](script)

def TTF_CloseFont(font):
    """Closes and frees the memory associated with a given font.

    This function should be called on a font when you are done with it. A
    :obj:`TTF_Font` cannot be used after it has been closed.
   
    Args:
        font (:obj:`TTF_Font`): The font to close.

    """
    return _funcs["TTF_CloseFont"](font)

def TTF_Quit():
    """De-initializes the TTF engine.
   
    Once this has been called, other functions in the module should not be used
    until :func:`TTF_Init` is called to re-initialize the engine (except for
    :func:`TTF_WasInit`).

    .. note::
       If :func:`TTF_Init` is called multiple times, this function should be
       called an equal number of times to properly de-initialize the library.

    """
    return _funcs["TTF_Quit"]()

def TTF_WasInit():
    """Checks if the TTF engine is initialized.
   
    This function can be used before calling :func:`TTF_Init` to avoid
    initializing twice in a row, or to determine if need to call
    :func:`TTF_Quit`.
   
    Returns:
        int: The number of times :func:`TTF_Init` has been called without a
        corresponding :func:`TTF_Quit`.

    """
    return _funcs["TTF_WasInit"]()

def TTF_GetFontKerningSize(font, prev_index, index):
    # NOTE: Deprecated in SDL_ttf
    return _funcs["TTF_GetFontKerningSize"](font, prev_index, index)

def TTF_GetFontKerningSizeGlyphs(font, previous_ch, ch):
    """Gets the kerning size of two glyphs (by FreeType index) for a given font.

    .. note::
       The units of kerning size returned by this function differ between fonts
       depending on their format and how they were designed.

    `Note: Added in SDL_ttf 2.0.14`

    Args:
        font (:obj:`TTF_Font`): The font object for which the kerning size
            should be retrieved.
        previous_ch (int): The unicode integer representing the first glyph.
        ch (int): The unicode integer representing the second glyph.

    Returns:
        int: The kerning size of the two glyphs in the current font.

    """
    return _funcs["TTF_GetFontKerningSizeGlyphs"](font, previous_ch, ch)

def TTF_GetFontKerningSizeGlyphs32(font, previous_ch, ch):
    """Gets the kerning size of two glyphs (by FreeType index) for a given font.

    Functionally identical to :func:`TTF_GetFontKerningSizeGlyphs`, except it
    supports 32-bit character codes instead of just 16-bit ones.

    `Note: Added in SDL_ttf 2.0.18`

    """
    return _funcs["TTF_GetFontKerningSizeGlyphs32"](font, previous_ch, ch)

def TTF_SetFontSDF(font, on_off):
    """Enables or disables Signed Distance Field rendering for a given font.

    Requires a version of FreeType that supports SDF rendering (2.11.0 or
    newer). As of SDL2_ttf 2.0.18, the FreeType version bundled with the
    official binaries is too old to support SDF.

    `Note: Added in SDL_ttf 2.0.18`

    Args:
        font (:obj:`TTF_Font`): The font object for which SDF should be
            enabled or disabled.
        on_off (int): Whether to enable (``SDL_TRUE``) or disable
            (``SDL_FALSE``) SDF rendering for the given font.

    Returns:
        int: 0 on success, or -1 if SDF support not available.

    """
    return _funcs["TTF_SetFontSDF"](font, on_off)

def TTF_GetFontSDF(font):
    """Checks if Signed Distance Field rendering is enabled for a given font.

    If using a version of FreeType without SDF support, this will always
    return 0.

    `Note: Added in SDL_ttf 2.0.18`

    Args:
        font (:obj:`TTF_Font`): The font object for which SDF usage should
            be checked.

    Returns:
        int: 1 if the font is using SDF, otherwise 0.

    """
    return _funcs["TTF_GetFontSDF"](font)

TTF_SetError = SDL_SetError
TTF_GetError = SDL_GetError



import os
import sys
import pytest
from sdl2 import ext as sdl2ext
from sdl2.ext import color
from sdl2.ext.surface import _create_surface
from sdl2 import surface, pixels

try:
    import numpy
    _HASNUMPY = True
except:
    _HASNUMPY = False


colors = {
    'red': color.Color(255, 0, 0, 255),
    'blue': color.Color(0, 0, 255, 255),
    'black': color.Color(0, 0, 0, 255),
    'white': color.Color(255, 255, 255, 255)
}

@pytest.fixture
def imgsurf(with_sdl):
    parent_dir = os.path.dirname(os.path.abspath(__file__))
    testfile = os.path.join(parent_dir, "resources", "surfacetest.bmp")
    sf = surface.SDL_LoadBMP(testfile.encode("utf-8"))
    yield sf
    surface.SDL_FreeSurface(sf)

@pytest.fixture
def rgbasurf(imgsurf):
    rgba = pixels.SDL_PIXELFORMAT_ABGR8888
    rgbasurf = surface.SDL_ConvertSurfaceFormat(imgsurf.contents, rgba, 0)
    yield rgbasurf
    surface.SDL_FreeSurface(rgbasurf)

@pytest.fixture
def surf_rgb24(with_sdl):
    grey = [128, 128, 128, 255]
    sf = _create_surface((16, 16), grey, fmt="RGB24")
    yield sf
    surface.SDL_FreeSurface(sf)


def test_PixelView(imgsurf, surf_rgb24):
    # Open pixel view for test image
    pxview = sdl2ext.PixelView(imgsurf.contents)

    # Test different coordinates on surface
    assert color.ARGB(pxview[0][0]) == colors['red']
    assert color.ARGB(pxview[0][16]) == colors['blue']
    assert color.ARGB(pxview[0][31]) == colors['white']
    assert color.ARGB(pxview[31][31]) == colors['black']

    # Try modifying surface, test if changes persist 
    pxview[31][0] = 0xFF808080 # medium grey in ARGB
    pxview2 = sdl2ext.PixelView(imgsurf)
    assert pxview2[31][0] == 0xFF808080

    # Test that negative indexing works as expected
    assert color.ARGB(pxview[0][-1]) == colors['white']
    assert color.ARGB(pxview[-1][-1]) == colors['black']

    # Test out-of-bounds exceptions for indices
    with pytest.raises(IndexError):
        pxview[32][32]

    # Test exception on use with a 3 byte per pixel surface
    with pytest.raises(RuntimeError):
        sdl2ext.PixelView(surf_rgb24)


@pytest.mark.skipif(not _HASNUMPY, reason="numpy module is not supported")
def test_pixels2d(imgsurf, surf_rgb24):
    # Open pixels2d view for test image
    nparray = sdl2ext.pixels2d(imgsurf.contents, transpose=False)
    assert nparray.shape == (32, 32)

    # Test different coordinates on surface
    assert color.ARGB(nparray[0][0]) == colors['red']
    assert color.ARGB(nparray[0][16]) == colors['blue']
    assert color.ARGB(nparray[0][31]) == colors['white']
    assert color.ARGB(nparray[31][31]) == colors['black']

    # Try modifying surface, test if changes persist 
    nparray[31][0] = 0xFF808080 # medium grey in ARGB
    nparray2 = sdl2ext.pixels2d(imgsurf, transpose=False)
    assert nparray2[31][0] == 0xFF808080

    # Test exception on use with a 3 byte per pixel surface
    with pytest.raises(RuntimeError):
        sdl2ext.pixels2d(surf_rgb24)


@pytest.mark.skipif(not _HASNUMPY, reason="numpy module is not supported")
def test_pixels3d(rgbasurf, surf_rgb24):
    # Create view and test different coordinates on surface
    nparray = sdl2ext.pixels3d(rgbasurf.contents, transpose=False)
    assert nparray.shape == (32, 32, 4)
    assert color.Color(*nparray[0][0]) == colors['red']
    assert color.Color(*nparray[0][16]) == colors['blue']
    assert color.Color(*nparray[0][31]) == colors['white']
    assert color.Color(*nparray[31][31]) == colors['black']

    # Create transposed view and test different coordinates on surface
    nptrans = sdl2ext.pixels3d(rgbasurf.contents, transpose=True)
    assert nptrans.shape == (32, 32, 4)
    assert color.Color(*nptrans[0][0]) == colors['red']
    assert color.Color(*nptrans[16][0]) == colors['blue']
    assert color.Color(*nptrans[31][0]) == colors['white']
    assert color.Color(*nptrans[31][31]) == colors['black']

    # Try modifying surface, test if changes persist
    grey = [128, 128, 128, 255]
    nparray[31][0][:] = grey
    nparray2 = sdl2ext.pixels3d(rgbasurf, transpose=False)
    assert color.Color(*nparray2[31][0]) == color.Color(*grey)

    # Test usage with a 3 bytes-per-pixel surface 
    nparray_rgb24 = sdl2ext.pixels3d(surf_rgb24)
    assert color.Color(*nparray_rgb24[0][0]) == color.Color(*grey)


@pytest.mark.skipif(not _HASNUMPY, reason="numpy module is not supported")
def test_surface_to_ndarray(imgsurf, rgbasurf):
    # Create a 2D ndarray from the surface & test different coordinates
    arr_2d = sdl2ext.surface_to_ndarray(imgsurf.contents, ndim=2)
    assert color.ARGB(arr_2d[0][0]) == colors['red']
    assert color.ARGB(arr_2d[0][16]) == colors['blue']
    assert color.ARGB(arr_2d[0][31]) == colors['white']
    assert color.ARGB(arr_2d[31][31]) == colors['black']

    # Create a 3D ndarray from the surface & test different coordinates
    arr_3d = sdl2ext.surface_to_ndarray(rgbasurf.contents)
    assert arr_3d.shape == (32, 32, 4)
    assert color.Color(*arr_3d[0][0]) == colors['red']
    assert color.Color(*arr_3d[0][16]) == colors['blue']
    assert color.Color(*arr_3d[0][31]) == colors['white']
    assert color.Color(*arr_3d[31][31]) == colors['black']

    # Try modifying surface, make sure changes don't persist
    grey = [128, 128, 128, 255]
    arr_3d[31][0][:] = grey
    arr_view = sdl2ext.pixels3d(rgbasurf, transpose=False)
    assert color.Color(*arr_view[31][0]) != color.Color(*grey)



from ctypes import POINTER, c_char_p, c_int, c_char
from .dll import _bind

__all__ = [
    # Enums
    "SDL_errorcode",
    "SDL_ENOMEM", "SDL_EFREAD", "SDL_EFWRITE", "SDL_EFSEEK",
    "SDL_UNSUPPORTED", "SDL_LASTERROR",

    # Macro Functions
    "SDL_OutOfMemory", "SDL_Unsupported", "SDL_InvalidParamError", "SDL_Error",

    # Functions
    "SDL_SetError", "SDL_GetError", "SDL_GetErrorMsg", "SDL_ClearError"
]


SDL_SetError = _bind("SDL_SetError", [c_char_p], c_int)
SDL_GetError = _bind("SDL_GetError", None, c_char_p)
SDL_GetErrorMsg = _bind("SDL_GetErrorMsg", [POINTER(c_char), c_int], c_char_p, added='2.0.14')
SDL_ClearError = _bind("SDL_ClearError")


SDL_errorcode = c_int

SDL_ENOMEM = 0
SDL_EFREAD = 1
SDL_EFWRITE = 2
SDL_EFSEEK = 3
SDL_UNSUPPORTED = 4
SDL_LASTERROR = 5


SDL_Error = _bind("SDL_Error", [SDL_errorcode], c_int)
SDL_OutOfMemory = SDL_Error(SDL_ENOMEM)
SDL_Unsupported = SDL_Error(SDL_UNSUPPORTED)
SDL_InvalidParamError = lambda x: SDL_SetError("Parameter '%s' is invalid" % (x))



import ctypes
from .compat import UnsupportedError, experimental
from .array import MemoryView
from ..surface import SDL_MUSTLOCK, SDL_LockSurface, SDL_UnlockSurface, \
    SDL_Surface
from ..stdinc import Uint8
from .draw import prepare_color
from .sprite import SoftwareSprite
from .surface import _get_target_surface

try:
    import numpy
    _HASNUMPY = True
except ImportError:
    _HASNUMPY = False


__all__ = [
    "PixelView", "SurfaceArray", "pixels2d", "pixels3d", "surface_to_ndarray"
]

class PixelView(MemoryView):
    """A 2D memory view for reading and writing SDL surface pixels.

    This class uses a ``view[y][x]`` layout, with the y-axis as the first
    dimension and the x-axis as the second. ``PixelView`` objects currently do
    not support array slicing, but support negative indexing as of
    PySDL2 0.9.10.

    If the source surface is RLE-accelerated, it will be locked automatically
    when the view is created and you will need to re-lock the surface using
    :func:`SDL_UnlockSurface` once you are done with the view.

    .. warning::
       The source surface should not be freed or deleted until the view is no
       longer needed. Accessing the view for a freed surface will likely cause
       Python to hard-crash.

    .. note:: 
       This class is implemented on top of the :class:`~sdl2.ext.MemoryView`
       class. As such, it makes heavy use of recursion to access rows and
       will generally be much slower than the :mod:`numpy`-based
       :func:`~sdl2.ext.pixels2d` and :func:`~sdl2.ext.pixels3d` functions.

    Args:
        source (:obj:`~sdl2.SDL_Surface`, :obj:`~sdl2.ext.SoftwareSprite`): The
            SDL surface to access with the view.

    """
    def __init__(self, source):
        if isinstance(source, SoftwareSprite):
            self._surface = source.surface
            # keep a reference, so the Sprite's not GC'd
            self._sprite = source
        elif isinstance(source, SDL_Surface):
            self._surface = source
        elif "SDL_Surface" in str(type(source)):
            self._surface = source.contents
        else:
            raise TypeError("source must be a Sprite or SDL_Surface")

        itemsize = self._surface.format.contents.BytesPerPixel
        if itemsize == 3:
            e = "Cannot open a 3 bytes-per-pixel surface using a PixelView."
            raise RuntimeError(e)

        if SDL_MUSTLOCK(self._surface):
            SDL_LockSurface(self._surface)

        pxbuf = ctypes.cast(self._surface.pixels, ctypes.POINTER(Uint8))
        strides = (self._surface.h, self._surface.w)
        srcsize = self._surface.h * self._surface.pitch
        super(PixelView, self).__init__(pxbuf, itemsize, strides,
                                        getfunc=self._getitem,
                                        setfunc=self._setitem,
                                        srcsize=srcsize)

    def _getitem(self, start, end):
        if self.itemsize == 1:
            # byte-wise access
            return self.source[start:end]
        # move the pointer to the correct location
        src = ctypes.byref(self.source.contents, start)
        casttype = ctypes.c_ubyte
        if self.itemsize == 2:
            casttype = ctypes.c_ushort
        elif self.itemsize == 4:
            casttype = ctypes.c_uint
        return ctypes.cast(src, ctypes.POINTER(casttype)).contents.value

    def _setitem(self, start, end, value):
        target = None
        if self.itemsize == 1:
            target = ctypes.cast(self.source, ctypes.POINTER(ctypes.c_ubyte))
        elif self.itemsize == 2:
            target = ctypes.cast(self.source, ctypes.POINTER(ctypes.c_ushort))
        elif self.itemsize == 4:
            target = ctypes.cast(self.source, ctypes.POINTER(ctypes.c_uint))
        value = prepare_color(value, self._surface)
        target[start // self.itemsize] = value


def _ndarray_prep(source, funcname, ndim):
    # Internal function for preparing SDL_Surfaces for casting to ndarrays
    if not _HASNUMPY:
        err = "'{0}' requires Numpy, which could not be found."
        raise UnsupportedError(err.format(funcname))

    # Get SDL surface and extract required attributes
    psurface = _get_target_surface(source, argname="source")
    sz = psurface.h * psurface.pitch
    bpp = psurface.format.contents.BytesPerPixel
    if bpp < 1 or bpp > 4:
        err = "The bpp of the source surface must be between 1 and 4, inclusive"
        raise ValueError(err + " (got {0}).".format(bpp))
    elif bpp == 3 and ndim == 2:
        err = "Surfaces with 3 bytes-per-pixel cannot be cast as 2D arrays."
        raise RuntimeError(err)

    # Handle 2D and 3D arrays differently where needed
    if ndim == 2:
        dtypes = {
            1: numpy.uint8,
            2: numpy.uint16,
            4: numpy.uint32
        }
        strides = (psurface.pitch, bpp)
        shape = psurface.h, psurface.w
        dtype = dtypes[bpp]
    else:
        strides = (psurface.pitch, bpp, 1)
        shape = psurface.h, psurface.w, bpp
        dtype = numpy.uint8

    return (psurface, sz, shape, dtype, strides)


def pixels2d(source, transpose=True):
    """Creates a 2D Numpy array view for a given SDL surface.

    This function casts the surface pixels to a 2D Numpy array view, providing
    read and write access to the underlying surface. If the source surface is
    RLE-accelerated, it will be locked automatically when the view is created
    and you will need to re-lock the surface using :func:`SDL_UnlockSurface`
    once you are done with the array.

    By default, the array is returned in ``arr[x][y]`` format with the x-axis
    as the first dimension, contrary to PIL and PyOpenGL convention. To obtain 
    an ``arr[y][x]`` array, set the ``transpose`` argument to ``False``.

    .. warning::
       The source surface should not be freed or deleted until the array is no
       longer needed. Accessing the array for a freed surface will likely cause
       Python to hard-crash.

    .. note::
       This function requires Numpy to be installed in the current Python
       environment.

    Args:
        source (:obj:`~sdl2.SDL_Surface`, :obj:`~sdl2.ext.SoftwareSprite`): The
            SDL surface to cast to a numpy array.
        transpose (bool, optional): Whether the output array should be
            transposed to have ``arr[x][y]`` axes instead of ``arr[y][x]`` axes.
            Defaults to ``True``.

    Returns:
        :obj:`numpy.ndarray`: A 2-dimensional Numpy array containing the integer
        color values for each pixel in the surface.

    Raises:
        RuntimeError: If Numpy could not be imported.
   
    """
    sf, sz, shape, dtype, strides = _ndarray_prep(source, "pixels2d", ndim=2)
    if SDL_MUSTLOCK(sf):
        SDL_LockSurface(sf)

    pxbuf = ctypes.cast(sf.pixels, ctypes.POINTER(ctypes.c_ubyte * sz))
    arr = SurfaceArray(
        shape, dtype, pxbuf.contents, 0, strides, "C", source, sf
    )
    return arr.transpose() if transpose else arr


def pixels3d(source, transpose=True):
    """Creates a 3D Numpy array view for a given SDL surface.

    This function casts the surface pixels to a 3D Numpy array view, providing
    read and write access to the underlying surface. If the source surface is
    RLE-accelerated, it will be locked automatically when the view is created
    and you will need to re-lock the surface using :func:`SDL_UnlockSurface`
    once you are done with the array.

    By default, the array is returned in ``arr[x][y]`` format with the x-axis
    as the first dimension, contrary to PIL and PyOpenGL convention. To obtain 
    an ``arr[y][x]`` array, set the ``transpose`` argument to ``False``.

    When creating a 3D array view, the order of the RGBA values for each pixel
    may be reversed for some common surface pixel formats (e.g. 'BGRA' for a
    ``SDL_PIXELFORMAT_ARGB8888`` surface). To correct this, you can call
    ``numpy.flip(arr, axis=2)`` to return a view of the array with the expected
    channel order.

    .. warning::
       The source surface should not be freed or deleted until the array is no
       longer needed. Accessing the array for a freed surface will likely cause
       Python to hard-crash.

    .. note::
       This function requires Numpy to be installed in the current Python
       environment.

    Args:
        source (:obj:`~sdl2.SDL_Surface`, :obj:`~sdl2.ext.SoftwareSprite`): The
            SDL surface to cast to a numpy array.
        transpose (bool, optional): Whether the output array should be
            transposed to have ``arr[x][y]`` axes instead of ``arr[y][x]`` axes.
            Defaults to ``True``.

    Returns:
        :obj:`numpy.ndarray`: A 3-dimensional Numpy array containing the values
        of each byte for each pixel in the surface.

    Raises:
        RuntimeError: If Numpy could not be imported.
   
    """
    sf, sz, shape, dtype, strides = _ndarray_prep(source, "pixels3d", ndim=3)
    if SDL_MUSTLOCK(sf):
        SDL_LockSurface(sf)

    pxbuf = ctypes.cast(sf.pixels, ctypes.POINTER(ctypes.c_ubyte * sz))
    arr = SurfaceArray(
        shape, dtype, pxbuf.contents, 0, strides, "C", source, sf
    )
    return arr.transpose(1, 0, 2) if transpose else arr


def surface_to_ndarray(source, ndim=3):
    """Returns a copy of an SDL surface as a Numpy array.
    
    The main difference between this function and :func:`~sdl2.ext.pixels2d` or
    :func:`~sdl2.ext.pixels3d` is that it returns a copy of the surface instead
    of a view, meaning that modifying the returned array will not affect the
    original surface (or vice-versa). This function is also slightly safer,
    as it does not assume that the source surface has been kept in memory.

    When creating a 3D array copy, the order of the RGBA values for each pixel
    may be reversed for some common surface pixel formats (e.g. 'BGRA' for a
    ``SDL_PIXELFORMAT_ARGB8888`` surface). To correct this, you can call
    ``numpy.flip(arr, axis=2)`` to return a view of the array with the expected
    channel order.

    .. note::
       Unlike :func:`~sdl2.ext.pixels2d` or :func:`~sdl2.ext.pixels3d`, this
       function always returns arrays with the y-axis as the first dimension
       (e.g. ``arr[y][x]``).

    .. note::
       This function requires Numpy to be installed in the current Python
       environment.

    Args:
        source (:obj:`~sdl2.SDL_Surface`, :obj:`~sdl2.ext.SoftwareSprite`): The
            SDL surface to convert to a numpy array.
        ndim (int, optional): The number of dimensions for the returned array,
            must be either 2 (for a 2D array) or 3 (for a 3D array). Defaults
            to 3.

    Returns:
        :obj:`numpy.ndarray`: A Numpy array containing a copy of the pixel data
        for the given surface.

    Raises:
        RuntimeError: If Numpy could not be imported.

    """
    if ndim not in [2, 3]:
        err = "Can only convert surfaces to 2D or 3D arrays (got {0})."
        raise ValueError(err.format(ndim))
    funcname = "surface_to_array"
    sf, sz, shape, dtype, strides = _ndarray_prep(source, funcname, ndim)
    was_unlocked = sf.locked == 0
    if SDL_MUSTLOCK(sf):
        SDL_LockSurface(sf)

    pxbuf = ctypes.cast(sf.pixels, ctypes.POINTER(ctypes.c_ubyte * sz))
    tmp = numpy.ndarray(shape, dtype, pxbuf.contents, strides=strides)
    if was_unlocked and SDL_MUSTLOCK(sf):
        SDL_UnlockSurface(sf)

    return numpy.copy(tmp)


class SurfaceArray(numpy.ndarray if _HASNUMPY else object):
    """A Numpy array that keeps a reference to its parent SDL surface.

    This class is used to keep track of the original source object for
    :func:`~sdl2.ext.pixels2d` or :func:`~sdl2.ext.pixels3d` to prevent it from
    being automatically freed during garbage collection. It should never be used
    for any other purpose.
    
    """
    def __new__(cls, shape, dtype=float, buffer_=None, offset=0,
                strides=None, order=None, source=None, surface=None):
        if _HASNUMPY:
            sfarray = numpy.ndarray.__new__(
                cls, shape, dtype, buffer_, offset, strides, order
            )
            sfarray._source = source
            sfarray._surface = surface
            return sfarray
        else:
            return None

    def __array_finalize__(self, sfarray):
        if sfarray is None:
            return
        self._source = getattr(sfarray, '_source', None)
        self._surface = getattr(sfarray, '_surface', None)



import sys
import pytest
from sdl2 import SDL_Init, SDL_Quit, SDL_INIT_VIDEO
from sdl2 import blendmode


@pytest.mark.skip("not implemented")
def test_SDL_ComposeCustomBlendMode(self, with_sdl):
    pass



import sys
import time
import pytest
import sdl2
from sdl2 import timer

if sys.version_info[0] >= 3:
    long = int


def test_SDL_GetTicks(with_sdl):
    ticks = timer.SDL_GetTicks()
    time.sleep(0.05)
    ticks2 = timer.SDL_GetTicks()
    time.sleep(0.05)
    ticks3 = timer.SDL_GetTicks()
    assert ticks2 > ticks
    assert ticks3 > ticks2

@pytest.mark.skipif(sdl2.dll.version < 2018, reason="not available")
def test_SDL_GetTicks64(with_sdl):
    ticks = timer.SDL_GetTicks64()
    time.sleep(0.05)
    ticks2 = timer.SDL_GetTicks64()
    time.sleep(0.05)
    ticks3 = timer.SDL_GetTicks64()
    assert ticks2 > ticks
    assert ticks3 > ticks2

def test_SDL_GetPerformanceCounter(with_sdl):
    perf = timer.SDL_GetPerformanceCounter()
    assert type(perf) in (int, long)
    assert perf > 0

def test_SDL_GetPerformanceFrequency(with_sdl):
    freq = timer.SDL_GetPerformanceFrequency()
    assert type(freq) in (int, long)
    assert freq > 0

@pytest.mark.xfail(reason="Unreliable on CI runners")
def test_SDL_Delay(with_sdl):
    for wait in [5, 10, 50, 100]:
        start = timer.SDL_GetTicks()
        timer.SDL_Delay(wait)
        end = timer.SDL_GetTicks()
        actual = (end - start)
        assert (wait - 2) <= actual <= (wait + 2)
                
@pytest.mark.skipif(hasattr(sys, "pypy_version_info"),
    reason="PyPy can't access other vars properly from a separate thread")
@pytest.mark.xfail(reason="Unreliable on CI runners")
def test_SDL_AddRemoveTimer(with_sdl):
    # Create a timer callback that adds a value to a Python list
    calls = []
    def timerfunc(interval, param):
        calls.append(param)
        return interval
    callback = timer.SDL_TimerCallback(timerfunc)
    timerid = timer.SDL_AddTimer(50, callback, "Test")
    # Run a loop for 300 ms and make sure the callback runs 5 or 6 times
    start = timer.SDL_GetTicks()
    while (timer.SDL_GetTicks() - start) <= 300:
        timer.SDL_Delay(10)
    assert len(calls) in [5, 6]
    # Try removing the timer and make sure the callback doesn't run anymore
    timer.SDL_RemoveTimer(timerid)
    timer.SDL_Delay(10)
    orig_calls = len(calls)
    start = timer.SDL_GetTicks()
    while (timer.SDL_GetTicks() - start) <= 200:
        timer.SDL_Delay(10)
    assert len(calls) == orig_calls
    # Wait a bit, so the last executing handlers can finish
    timer.SDL_Delay(10)



import pytest
from sdl2.ext import algorithms as alg


def test_cohensutherland():
    # Test a simple diagonal intersection
    ret = alg.cohensutherland(1, 2, 2, 1, 0, 0, 4, 4)
    assert ret == (1, 1, 2, 2)

    # Test handling of swapped top/bottom values
    ret = alg.cohensutherland(1, 1, 2, 2, 0, 0, 4, 4)
    assert ret == (1, 1, 2, 2)

    # Test partial intersection
    ret = alg.cohensutherland(1, 1, 2, 2, 1.5, 0, 1.5, 1.5)
    assert ret == (1.5, 1, 1.5, 1.5)

    # Test non-intersecting line
    ret = alg.cohensutherland(1, 2, 2, 1, 0, 0, 0, 4)
    assert ret == (None, None, None, None)


def test_liangbarsky():
    # Test a simple diagonal intersection
    ret = alg.liangbarsky(1, 2, 2, 1, 0, 0, 4, 4)
    assert ret == (1, 1, 2, 2)

    # Test handling of swapped top/bottom values
    ret = alg.liangbarsky(1, 1, 2, 2, 0, 0, 4, 4)
    assert ret == (1, 1, 2, 2)

    # Test partial intersection
    ret = alg.liangbarsky(1, 1, 2, 2, 1.5, 0, 1.5, 1.5)
    assert ret == (1.5, 1, 1.5, 1.5)

    # Test non-intersecting line
    ret = alg.liangbarsky(1, 2, 2, 1, 0, 0, 0, 4)
    assert ret == (None, None, None, None)


def test_clipline():
    # Test a simple diagonal intersection w/ default method
    ret = alg.clipline(1, 2, 2, 1, 0, 0, 4, 4)
    assert ret == (1, 1, 2, 2)

    # Test a simple diagonal intersection w/ Liang-Barsky
    ret = alg.clipline(1, 2, 2, 1, 0, 0, 4, 4, method='liangbarsky')
    assert ret == (1, 1, 2, 2)

    # Test a simple diagonal intersection w/ Cohen-Sutherland
    ret = alg.clipline(1, 2, 2, 1, 0, 0, 4, 4, method='cohensutherland')
    assert ret == (1, 1, 2, 2)

    # Test exception with unknown clipping method
    with pytest.raises(ValueError):
        alg.clipline(1, 2, 2, 1, 0, 0, 4, 4, method='cohenlebowski')


def test_point_on_line():
    line = [(1, 1), (3, 3)]
    assert alg.point_on_line(line[0], line[1], (2, 2))
    assert alg.point_on_line(line[0], line[1], (3, 3))
    assert alg.point_on_line(line[1], line[0], (2, 2))
    assert not alg.point_on_line(line[0], line[1], (3, 4))
